"""
Copyright (C) 2024-2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List

import numpy as np
from onnx.helper import tensor_dtype_to_np_dtype
from onnx.onnx_pb import NodeProto

from ... import logger
from ...graph import OnnxGraph
from .. import PASSES
from ..pattern import SingleNodePattern
from ..rewriter import Rewriter
from ..utils import make_constant


@PASSES.register(
    "canonicalize_negative_zero_point", deps=["canonicalize_negative_scale"]
)
class CanonicalizeNegativeZeroPointRewriter(Rewriter):
    """Negative zero point is not supported on some devices.
    For negative zero point, we should increase the scale to enlarge the range
    of the data.

    Before:

        x [1, 5] (float32) / 0.01569 - 64 = qx [0, 255]
        (scale [0.01569], zero_point [-64])

    After:

        x [1, 5] (float32) / 0.01960 - 0 = qx [0, 255]
        (scale [0.01960], zero_point [0])
    """

    def __init__(self):
        super().__init__(pattern=SingleNodePattern("DequantizeLinear"))

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto]):
        node = nodes[0]
        scale = self.get_value(node.input[1])
        zero_point = self.get_value(node.input[2])
        if zero_point is None or np.all(zero_point >= 0):
            return
        if scale is None:
            raise ValueError(
                f"encounter node {node.name} with negative zero point but "
                "nonconst scale"
            )
        qtype = tensor_dtype_to_np_dtype(graph.tensor_type(node.input[2]))
        bits = np.iinfo(qtype).bits
        qmax = 2**bits - 1
        x_max = (qmax - zero_point.astype("int32")) * scale
        x_min = np.minimum((0 - zero_point) * scale, 0)  # x_min should below zero
        new_scale = (x_max - x_min) / qmax
        new_zero_point = np.maximum(zero_point, 0)
        logger.warning(
            f"change {node.name}: scale {scale} -> {new_scale}, "
            f"zero_point {zero_point} -> {new_zero_point}"
        )

        scale_cst = make_constant(f"{node.name}/scale", new_scale)
        zp_cst = make_constant(f"{node.name}/zero_point", new_zero_point)

        node.input[1] = scale_cst.output[0]
        node.input[2] = zp_cst.output[0]

        if prev_n := self.get_input_node(node, 0):
            if prev_n.op_type in ("QuantizeLinear",):
                prev_n.input[1] = scale_cst.output[0]
                prev_n.input[2] = zp_cst.output[0]

        self += [scale_cst, zp_cst]



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# check onnxscript and torch installation

try:
    import torch

    if torch.__version__ < "2.0.0":
        raise ImportError("PyTorch version should be 2.0.0 or higher.")
except ImportError as e:
    raise ImportError("PyTorch>=2.0.0 is not installed.") from e


try:
    import onnxscript
except ImportError as e:
    raise ImportError("onnxscript is not installed.") from e


from onnx import OperatorSetIdProto
from onnx.helper import make_operatorsetid

IR_DOMAIN: OperatorSetIdProto = make_operatorsetid("ai.pytorch.aten", 2)
"""Pytorch ATen"""



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=missing-function-docstring

import random

import onnx
import pytest

from onnxifier import OnnxGraph, PassManager
from onnxifier.passes import PASSES, Rewriter
from onnxifier.passes.pattern import SingleNodePattern


@PASSES.register(name="function_in_test")
def fake_pass_function(graph, args_x, args_y):
    assert args_x == args_y
    return graph


@PASSES.register(name="class_in_test")
class FakeClassPass(Rewriter):
    """Fake Pass"""

    def __init__(self):
        super().__init__(SingleNodePattern())

    def rewrite(self, graph, nodes, args_x=0, args_y=2):
        assert args_x == args_y


@PASSES.register("cycle_a", deps=["cycle_b"])
def fake_cycle_a(graph):
    return graph


@PASSES.register("cycle_b", deps=["cycle_a"])
def fake_cycle_b(graph):
    return graph


@PASSES.register("patch_a", patch=["patch_b"])
def fake_patch_a(graph):
    return graph


@PASSES.register("patch_b", patch=["patch_a"])
def fake_patch_b(graph):
    return graph


def _empty_model():
    return onnx.helper.make_model(
        graph=onnx.helper.make_graph([], "empty", [], []),
    )


def test_pass_manager_default():
    graph = OnnxGraph(_empty_model())
    pass_manager = PassManager()
    pass_manager.optimize(graph)


def test_pass_manager_include_and_exclude():
    passes = list(iter(PASSES))
    random.shuffle(passes)
    cut_pos = len(passes) // 2
    pass_manager = PassManager(passes[:cut_pos], passes[cut_pos:])
    graph = OnnxGraph(_empty_model())
    pass_manager.optimize(graph)


def test_pass_manager_include_instance():
    pass_manager = PassManager([FakeClassPass()])
    graph = OnnxGraph(_empty_model())
    pass_manager.optimize(graph)


def test_pass_manager_include_warnings(caplog):
    PassManager(["not_exist_pass"])
    assert "WARNING" in caplog.text


def test_pass_manager_with_configs():
    pass_manager = PassManager(
        ["function_in_test", "class_in_test", "class_in_test", "function_in_test"],
        configs={
            "function_in_test": dict(args_x=1, args_y=1),
            "class_in_test:0": dict(args_x=1, args_y=1),
            "class_in_test:1": dict(args_x="2", args_y="2"),
        },
    )
    graph = OnnxGraph(_empty_model())
    pass_manager.optimize(graph, strict=True)


def test_pass_manager_check_cycle():
    pm = PassManager(["cycle_a", "cycle_b"])
    with pytest.raises(RuntimeError):
        pm.optimize(OnnxGraph(_empty_model()), strict=True)

    pm = PassManager(["patch_a", "patch_b"])
    with pytest.raises(RuntimeError):
        pm.optimize(OnnxGraph(_empty_model()), strict=True)


def test_pass_child():
    assert "patch_b" in PASSES.child("patch_b")
    assert "patch_a" in PASSES.child(["patch_a", "patch_b"])
    assert "cycle_a" not in PASSES.child(["patch_a", "patch_b"])

    with pytest.raises(KeyError):
        PASSES.child("not_exist")


@PASSES.register("test_pass_recurse")
class PassRecurse(Rewriter):
    """Fake Pass"""

    def __init__(self, debug_info=None):
        super().__init__(SingleNodePattern())
        self.debug_info = debug_info

    def rewrite(self, graph, nodes):
        self.debug_info[nodes[0].name] = nodes[0].op_type
        return graph


def test_optimize_recursively():
    m = _empty_model()
    m.graph.node.append(onnx.helper.make_node("Foo", [], [], name="foo", domain="foo"))
    m.functions.append(
        onnx.helper.make_function(
            "foo",
            "Foo",
            [],
            [],
            [onnx.helper.make_node("Const", [], ["out"], name="const")],
            [onnx.helper.make_operatorsetid("foo", 1)],
        )
    )
    g = OnnxGraph(m)

    debug = {}
    rewriter = PassRecurse(debug)
    pm = PassManager([rewriter])
    pm.optimize(g, recursive=False)

    assert debug["foo"] == "Foo"
    assert "const" not in debug

    pm.optimize(g, recursive=True)
    assert debug["const"] == "Const"


def test_optimize_recursively_with_nested_functions_out_of_order():
    m = _empty_model()
    m.graph.node.append(onnx.helper.make_node("Foo", [], [], name="foo", domain="foo"))
    m.functions.append(
        onnx.helper.make_function(
            "foo",
            "Foo",
            [],
            [],
            [onnx.helper.make_node("Bar", [], ["out"], name="bar_call")],
            [onnx.helper.make_operatorsetid("foo", 1)],
        )
    )
    m.functions.append(
        onnx.helper.make_function(
            "bar",
            "Bar",
            [],
            [],
            [onnx.helper.make_node("Const", [], ["out"], name="const")],
            [onnx.helper.make_operatorsetid("bar", 1)],
        )
    )
    g = OnnxGraph(m)

    debug = {}
    rewriter = PassRecurse(debug)
    pm = PassManager([rewriter])
    pm.optimize(g, recursive=True, strict=True)

    assert debug["foo"] == "Foo"
    assert debug["bar_call"] == "Bar"
    assert debug["const"] == "Const"


def test_optimize_with_specify_node_names():
    m = _empty_model()
    m.graph.node.append(onnx.helper.make_node("Foo", [], [], name="foo", domain="foo"))
    m.graph.node.append(onnx.helper.make_node("Bar", [], [], name="bar", domain="bar"))
    g = OnnxGraph(m)

    pm = PassManager(
        ["function_in_test"], configs={"function_in_test": {"args_x": 1, "args_y": 2}}
    )
    # This function_in_test should fail since args_x != args_y
    with pytest.raises(AssertionError):
        pm.optimize(g, strict=True, specify_node_names={"foo"})
    with pytest.raises(AssertionError):
        # specify_node_names has no effect to function passes
        pm.optimize(g, strict=True, specify_node_names={"baz"})

    pm = PassManager(
        ["class_in_test"], configs={"class_in_test": {"args_x": 2, "args_y": 3}}
    )
    # This class_in_test should fail since args_x != args_y
    with pytest.raises(AssertionError):
        pm.optimize(g, strict=True, specify_node_names={"foo"})
    # However, specify node name = "baz" matches no nodes
    pm.optimize(g, strict=True, specify_node_names={"baz"})



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List

from onnx.onnx_pb import NodeProto

from ... import OnnxGraph, logger
from .. import PASSES
from ..pattern import SingleNodePattern
from ..rewriter import Rewriter


@PASSES.register("eliminate_reshape", deps=["infer_shape"])
class EliminateInputOutputReshapeRewriter(Rewriter):
    """Eliminate useless reshape operations."""

    def __init__(self):
        super().__init__(SingleNodePattern("Reshape"))

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto], *args, **kwargs):
        node = nodes[0]
        self._rewrite_input_reshape(graph, node)
        self._rewrite_output_reshape(graph, node)

    def _rewrite_input_reshape(self, graph: OnnxGraph, node: NodeProto):
        """remove the reshape if it is the input of the graph

        Before:
            input(shape1) -> reshape(shape2) -> conv

        After:
            input(shape2) -> conv
        """
        if node.input[0] in graph.inputs:
            input_name = node.input[0]
            graph.remove_input(input_name)
            for n in graph.onnx_successors(node):
                graph.set_input(n, node.output[0])
                input_pos = [i for i in n.input].index(node.output[0])
                logger.debug(f"set input {input_pos} of {n.name} to {input_name}")
                graph.rename_input(node.output[0], input_name)
                n.input[input_pos] = input_name
            self -= node

    def _rewrite_output_reshape(self, graph: OnnxGraph, node: NodeProto):
        """remove the reshape if it is the output of the graph

        Before:
            conv(shape1) -> reshape(shape2) -> output(shape2)

        After:
            conv -> output(shape1)
        """
        if node.output[0] in graph.outputs:
            output_name = node.output[0]
            graph.remove_output(output_name)
            for n in graph.onnx_predecessors(node):
                graph.set_output(n, node.input[0])
                output_pos = [i for i in n.output].index(node.input[0])
                logger.debug(f"set output {output_pos} of {n.name} to {output_name}")
                graph.rename_output(node.input[0], output_name)
                n.output[output_pos] = output_name
            self -= node



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=arguments-differ

from typing import List

from onnx.helper import make_node
from onnx.onnx_pb import NodeProto

from ... import OnnxGraph
from .. import L1
from ..pattern import SingleNodePattern
from ..rewriter import Rewriter


@L1.register(name="prelu_to_leaky", deps=["initializer_to_constant"])
class PReluToLeakyRewriter(Rewriter):
    """Convert PRelu whose parameter is a scaler to LeakyRelu."""

    def __init__(self):
        super().__init__(pattern=SingleNodePattern("PRelu"))

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto], *args, **kwargs):
        node = nodes[0]
        # quick check slope shape
        slope_shape = graph.tensor_shape(node.input[1])
        if len(slope_shape) >= 1 and slope_shape[0] != 1:
            # keep PRelu
            return

        slope_value = None
        for pred in graph.onnx_predecessors(node):
            if node.input[1] in pred.output:
                slope_value = self.get_value(pred)
                break
        if slope_value is None:
            # slope is not static
            return

        leakyrelu_node = make_node(
            "LeakyRelu",
            inputs=node.input[:1],
            outputs=node.output,
            name=node.name + "/to_leaky",
            alpha=float(slope_value.squeeze()),
        )
        self += leakyrelu_node
        self -= node



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import onnx
from onnx.helper import (
    make_graph,
    make_model,
    make_node,
    make_tensor,
    make_tensor_type_proto,
    make_value_info,
)
from onnx.reference import ReferenceEvaluator

from onnxifier import PassManager
from onnxifier.graph import OnnxGraph


def _build_test_graph(alpha, beta, transA, transB):
    gemm = make_node(
        "Gemm",
        inputs=["x", "B", "C"],
        outputs=["y"],
        name="gemm",
        alpha=float(alpha),
        beta=float(beta),
        transA=transA,
        transB=transB,
    )
    x_shape = (64, 1) if transA else (1, 64)
    b_shape = (32, 64) if transB else (64, 32)
    graph = make_graph(
        [gemm],
        "graph",
        [make_value_info("x", make_tensor_type_proto(1, x_shape))],
        [make_value_info("y", make_tensor_type_proto(1, [1, 32]))],
        [
            make_tensor("B", 1, b_shape, np.random.rand(*b_shape)),
            make_tensor("C", 1, [32], np.random.rand(32)),
        ],
    )
    return make_model(graph)


def _build_quantize_test_graph(alpha, beta, transA, transB):
    gemm = make_node(
        "Gemm",
        inputs=["A", "B", "C"],
        outputs=["Y"],
        name="gemm",
        alpha=float(alpha),
        beta=float(beta),
        transA=transA,
        transB=transB,
    )
    dequant = make_node(
        "DequantizeLinear",
        inputs=["x", "x_scale", "x_zero_point"],
        outputs=["B"],
        name="dequantzie",
        axis=0 if transB else 1,
    )

    a_shape = (64, 1) if transA else (1, 64)
    b_shape = (32, 64) if transB else (64, 32)
    graph = make_graph(
        [dequant, gemm],
        "graph",
        [make_value_info("A", make_tensor_type_proto(1, a_shape))],
        [make_value_info("Y", make_tensor_type_proto(1, [1, 32]))],
        [
            make_tensor("C", 1, [32], np.random.rand(32)),
            make_tensor(
                "x", 3, b_shape, np.random.randint(-128, 128, b_shape, dtype="int8")
            ),
            make_tensor("x_scale", 1, [32], np.random.rand(32)),
            make_tensor(
                "x_zero_point",
                3,
                [32],
                np.random.randint(-128, 128, [32], dtype="int8"),
            ),
        ],
    )
    return make_model(graph)


def _build_test_ND_graph(shape1, shape2):
    matmul = make_node(
        "MatMul",
        inputs=["x", "w"],
        outputs=["y"],
        name="matmul",
    )
    x_shape = shape1
    b_shape = shape2
    out_shape = np.maximum(shape1[:-2], shape2[:-2]).tolist() + [shape1[-2], shape2[-1]]
    graph = make_graph(
        [matmul],
        "graph",
        [
            make_value_info("x", make_tensor_type_proto(1, x_shape)),
            make_value_info("w", make_tensor_type_proto(1, b_shape)),
        ],
        [make_value_info("y", make_tensor_type_proto(1, out_shape))],
        [],
    )
    return make_model(graph)


def test_rewriter():
    graph = OnnxGraph(_build_test_graph(1, 1, 0, 0))
    pm = PassManager(["initializer_to_constant", "gemm_to_conv"])
    graph = pm.optimize(graph, strict=True)
    onnx.checker.check_model(graph.model, True)


def test_rewriter_transB():
    graph = OnnxGraph(_build_test_graph(1, 1, 0, 1))
    pm = PassManager(["initializer_to_constant", "gemm_to_conv"])
    graph = pm.optimize(graph, strict=True)
    onnx.checker.check_model(graph.model, True)


def test_rewriter_transA():
    graph = OnnxGraph(_build_test_graph(1, 1, 1, 0))
    pm = PassManager(["initializer_to_constant", "gemm_to_conv"])
    graph = pm.optimize(graph, strict=True)
    onnx.checker.check_model(graph.model, True)


def test_quantize_rewriter():
    graph = OnnxGraph(_build_quantize_test_graph(1, 1, 0, 0))
    pm = PassManager(["initializer_to_constant", "gemm_to_conv"])
    graph = pm.optimize(graph, strict=True)
    onnx.checker.check_model(graph.model, True)


def test_quantize_rewriter_transA():
    graph = OnnxGraph(_build_quantize_test_graph(1, 1, 1, 0))
    pm = PassManager(["initializer_to_constant", "gemm_to_conv"])
    graph = pm.optimize(graph, strict=True)
    onnx.checker.check_model(graph.model, True)


def test_quantize_rewriter_transB():
    graph = OnnxGraph(_build_quantize_test_graph(1, 1, 0, 1))
    pm = PassManager(["initializer_to_constant", "gemm_to_conv"])
    graph = pm.optimize(graph, strict=True)
    onnx.checker.check_model(graph.model, True)


def test_quantize_rewriter_scale_alpha():
    graph = OnnxGraph(_build_quantize_test_graph(0.5, 1, 0, 1))
    pm = PassManager(["initializer_to_constant", "gemm_to_conv"])
    graph = pm.optimize(graph, strict=True)
    onnx.checker.check_model(graph.model, True)


def test_rewriter_3D():
    x_shape = [128, 4, 64]
    w_shape = [1, 64, 16]
    graph = OnnxGraph(_build_test_ND_graph(x_shape, w_shape))
    runner1 = ReferenceEvaluator(graph.model)
    x = np.random.rand(*x_shape).astype("float32")
    w = np.random.rand(*w_shape).astype("float32")
    y1 = runner1.run(None, {"x": x, "w": w})

    pm = PassManager(["initializer_to_constant", "gemm_to_conv"])
    graph = pm.optimize(graph, strict=True)
    onnx.checker.check_model(graph.model, True)

    runner2 = ReferenceEvaluator(graph.model)
    y2 = runner2.run(None, {"x": x, "w": w})
    assert np.allclose(y1, y2)


def test_rewriter_3_2D():
    x_shape = [1, 4, 64]
    w_shape = [64, 16]
    graph = OnnxGraph(_build_test_ND_graph(x_shape, w_shape))
    runner1 = ReferenceEvaluator(graph.model)
    x = np.random.rand(*x_shape).astype("float32")
    w = np.random.rand(*w_shape).astype("float32")
    y1 = runner1.run(None, {"x": x, "w": w})

    pm = PassManager(["initializer_to_constant", "gemm_to_conv"])
    graph = pm.optimize(graph, strict=True)
    onnx.checker.check_model(graph.model, True)

    runner2 = ReferenceEvaluator(graph.model)
    y2 = runner2.run(None, {"x": x, "w": w})
    assert np.allclose(y1, y2)


def test_rewriter_3ND_swap():
    x_shape = [1, 4, 64]
    w_shape = [128, 64, 16]
    graph = OnnxGraph(_build_test_ND_graph(x_shape, w_shape))
    runner1 = ReferenceEvaluator(graph.model)
    x = np.random.rand(*x_shape).astype("float32")
    w = np.random.rand(*w_shape).astype("float32")
    y1 = runner1.run(None, {"x": x, "w": w})

    pm = PassManager(["initializer_to_constant", "gemm_to_conv"])
    graph = pm.optimize(graph, strict=True)
    onnx.checker.check_model(graph.model, True)

    runner2 = ReferenceEvaluator(graph.model)
    y2 = runner2.run(None, {"x": x, "w": w})
    assert np.allclose(y1, y2)


def test_rewriter_4D():
    x_shape = [1, 128, 4, 64]
    w_shape = [1, 1, 64, 16]
    graph = OnnxGraph(_build_test_ND_graph(x_shape, w_shape))
    runner1 = ReferenceEvaluator(graph.model)
    x = np.random.rand(*x_shape).astype("float32")
    w = np.random.rand(*w_shape).astype("float32")
    y1 = runner1.run(None, {"x": x, "w": w})

    pm = PassManager(["initializer_to_constant", "gemm_to_conv"])
    graph = pm.optimize(graph, strict=True)
    onnx.checker.check_model(graph.model, True)

    runner2 = ReferenceEvaluator(graph.model)
    y2 = runner2.run(None, {"x": x, "w": w})
    assert np.allclose(y1, y2)


def test_rewriter_4ND_swap():
    x_shape = [1, 1, 4, 64]
    w_shape = [1, 128, 64, 16]
    graph = OnnxGraph(_build_test_ND_graph(x_shape, w_shape))
    runner1 = ReferenceEvaluator(graph.model)
    x = np.random.rand(*x_shape).astype("float32")
    w = np.random.rand(*w_shape).astype("float32")
    y1 = runner1.run(None, {"x": x, "w": w})

    pm = PassManager(["initializer_to_constant", "gemm_to_conv"])
    graph = pm.optimize(graph, strict=True)
    onnx.checker.check_model(graph.model, True)

    runner2 = ReferenceEvaluator(graph.model)
    y2 = runner2.run(None, {"x": x, "w": w})
    assert np.allclose(y1, y2)


def test_rewriter_5D():
    x_shape = [1, 1, 128, 4, 64]
    w_shape = [1, 1, 1, 64, 16]
    graph = OnnxGraph(_build_test_ND_graph(x_shape, w_shape))
    runner1 = ReferenceEvaluator(graph.model)
    x = np.random.rand(*x_shape).astype("float32")
    w = np.random.rand(*w_shape).astype("float32")
    y1 = runner1.run(None, {"x": x, "w": w})

    pm = PassManager(["initializer_to_constant", "gemm_to_conv"])
    graph = pm.optimize(graph, strict=True)
    onnx.checker.check_model(graph.model, True)

    runner2 = ReferenceEvaluator(graph.model)
    y2 = runner2.run(None, {"x": x, "w": w})
    assert np.allclose(y1, y2)


def test_rewriter_5ND_swap():
    x_shape = [1, 1, 1, 4, 64]
    w_shape = [1, 2, 128, 64, 16]
    graph = OnnxGraph(_build_test_ND_graph(x_shape, w_shape))
    runner1 = ReferenceEvaluator(graph.model)
    x = np.random.rand(*x_shape).astype("float32")
    w = np.random.rand(*w_shape).astype("float32")
    y1 = runner1.run(None, {"x": x, "w": w})

    pm = PassManager(["initializer_to_constant", "gemm_to_conv"])
    graph = pm.optimize(graph, strict=True)
    onnx.checker.check_model(graph.model, True)

    runner2 = ReferenceEvaluator(graph.model)
    y2 = runner2.run(None, {"x": x, "w": w})
    assert np.allclose(y1, y2)



"""
Copyright (C) 2023-2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import tempfile
from copy import deepcopy
from pathlib import Path
from typing import Dict, Iterable, Optional, Sequence, Set, Tuple

import networkx as nx
import numpy as np
import onnx
from onnx import AttributeProto, NodeProto, TensorProto, numpy_helper, save_model
from onnx.helper import (
    make_graph,
    make_model,
    make_node,
    make_tensor_type_proto,
    make_value_info,
)

from ..graph import OnnxGraph
from ..logger import debug


def make_constant(name: str, value: np.ndarray) -> "NodeProto":
    """Make a Constant node according to given value."""
    node = make_node(
        op_type="Constant",
        name=name,
        inputs=[],
        outputs=[f"{name}_output_0"],
        value=numpy_helper.from_array(value),
    )
    return node


def attribute_value(attr: AttributeProto):
    """Get the value of an onnx attribute."""
    match attr.type:
        case AttributeProto.FLOAT:
            return float(attr.f)
        case AttributeProto.INT:
            return int(attr.i)
        case AttributeProto.STRING:
            return attr.s.decode("utf-8")
        case AttributeProto.TENSOR:
            return numpy_helper.to_array(attr.t)
        case AttributeProto.TYPE_PROTO:
            return attr.tp
        case AttributeProto.FLOATS:
            return [float(f) for f in attr.floats]
        case AttributeProto.INTS:
            return [int(i) for i in attr.ints]
        case AttributeProto.STRINGS:
            return [s.decode("utf-8") for s in attr.strings]
        case AttributeProto.TENSORS:
            return [numpy_helper.to_array(t) for t in attr.tensors]
        case AttributeProto.GRAPHS:
            return [g for g in attr.graphs]
        case AttributeProto.TYPE_PROTOS:
            return [tp for tp in attr.type_protos]
    raise ValueError(f"Unsupported attribute type {attr.type}")


def canonical_node_name(name: str) -> str:
    """Canonicalize a node name by replacing illegal characters with '_'."""

    return name.replace(".", "_").replace(":", "_").replace("/", "_")


def is_elewise(node: NodeProto | str) -> bool:
    """Return True if the node type is an element-wise operation."""

    if isinstance(node, NodeProto):
        node = node.op_type
    return node in {
        "Abs",
        "Acos",
        "Acosh",
        "Add",
        "And",
        "Asin",
        "Asinh",
        "Atan",
        "Atanh",
        "BatchNormalization",
        "Bernoulli",
        "BitShift",
        "BitwiseAnd",
        "BitwiseNot",
        "BitwiseOr",
        "BitwiseXor",
        "Cast",
        "CastLike",
        "Ceil",
        "Celu",
        "Clip",
        "Cos",
        "Cosh",
        "CumSum",
        "Div",
        "Einsum",
        "Elu",
        "Equal",
        "Erf",
        "Exp",
        "Floor",
        "Gelu",
        "Greater",
        "GreaterOrEqual",
        "HardSigmoid",
        "HardSwish",
        "Hardmax",
        "Identity",
        "LeakyRelu",
        "Less",
        "LessOrEqual",
        "Log",
        "LogSoftmax",
        "LpNormalization",
        "Max",
        "Mean",
        "Min",
        "Mish",
        "Mod",
        "Mul",
        "Neg",
        "Not",
        "Or",
        "PRelu",
        "Pow",
        "Reciprocal",
        "Relu",
        "Round",
        "Selu",
        "Sigmoid",
        "Sign",
        "Sin",
        "Sinh",
        "Softmax",
        "Softplus",
        "Softsign",
        "Sqrt",
        "Sub",
        "Sum",
        "Tan",
        "Tanh",
        "Xor",
    }


def evaluate_on_node(
    graph, node: NodeProto, output_name: Optional[str] = None
) -> Optional[np.ndarray]:
    """Evaluate an output of a given node in the graph.

    Args:
        graph (OnnxGraph): the model graph.
        node (NodeProto): specify a node to evaluate its output.
        output_name (Optional[str], optional): specify the output name of the node.
            Defaults to None to use the first output of the node.

    Returns:
        Optional[np.ndarray]: a constant array value if succeed, None otherwise.
    """
    all_preds = nx.traversal.dfs_predecessors(graph.reverse(), node.name)
    all_preds = set(all_preds)
    all_preds.add(node.name)  # include self
    h = deepcopy(graph.onnx_subgraph(all_preds))
    try:
        # pylint: disable=import-outside-toplevel
        from onnxifier import convert_graph

        sub_model = h.model
        # extend current value_info due to ``h.model`` cleared it.
        # it's needed because shape inference can not work for custom domains.
        # pylint: disable=protected-access
        sub_model.graph.value_info.extend(h._value_info)
        sub_model = convert_graph(
            sub_model,
            [],
            print_passes=False,
            target_opset=graph.opset_version,
        ).model
    except Exception:  # pylint: disable=broad-except
        debug(f"convert subgraph failed on {node.name}: [{all_preds}]")
        return
    try:
        # pylint: disable=import-outside-toplevel
        from onnxifier.evaluator import Evaluator

        if output_name is None:
            output_name = sub_model.graph.output[0].name
        runner = Evaluator(sub_model)
        return runner([output_name], {})[0]
    except Exception:  # pylint: disable=broad-except
        node_name = canonical_node_name(node.name)
        temp_save = Path(tempfile.gettempdir()) / f"{node_name}.onnx"
        save_model(sub_model, temp_save)
        debug(f"evaluate subgraph failed on {node.name}: {temp_save}")
        return


def cast_in(node: NodeProto, index: int, to: int) -> NodeProto:
    """Insert a Cast node before the index-th input of the node.

    Args:
        node (NodeProto): specify a node to insert a Cast node.
        index (int): the index of the input to insert the Cast node.
        to (int): the data type of the Cast node.

    Note:
        You should pass in a `Rewriter` object if it is called inside a `rewrite`
        method.
    """

    if index < 0:
        index += len(node.input)
    assert index <= len(node.input)
    assert to != TensorProto.UNDEFINED
    cast_node = make_node(
        "Cast",
        inputs=[node.input[index]],
        outputs=[f"{node.name}/cast_i{index}_output0"],
        name=f"{node.name}/cast_i{index}",
        to=to,
    )
    node.input[index] = cast_node.output[0]
    return cast_node


def cast_out(node: NodeProto, index: int, to: int) -> NodeProto:
    """Insert a Cast node after the index-th output of the node.

    Args:
        node (NodeProto): specify a node to insert a Cast node.
        index (int): the index of the output to insert the Cast node.
        to (int): the data type of the Cast node.

    Note:
        You should pass in a `Rewriter` object if it is called inside a `rewrite`
        method.
    """

    if index < 0:
        index += len(node.output)
    assert index <= len(node.output)
    assert to != TensorProto.UNDEFINED
    cast_node = make_node(
        "Cast",
        inputs=[f"{node.name}/cast_o{index}_input0"],
        outputs=[node.output[index]],
        name=f"{node.name}/cast_o{index}",
        to=to,
    )
    node.output[index] = cast_node.input[0]
    return cast_node


def _populate_functions(
    nodes_type: Set[str], function_map: Dict[str, onnx.FunctionProto]
) -> Iterable[onnx.FunctionProto]:
    """find all used functions"""
    used_functions = {}
    while nodes_type:
        node_type = nodes_type.pop()
        if node_type not in function_map and node_type not in used_functions:
            continue
        used_functions[node_type] = function_map[node_type]
        nodes_type.update({node.op_type for node in function_map[node_type].node})
    return used_functions.values()


def extract_function(
    graph: OnnxGraph,
    nodes: Sequence[onnx.NodeProto],
    inputs: Sequence[Tuple[str, str]],
    outputs: Sequence[Tuple[str, str]],
) -> OnnxGraph:
    """Create a new onnx graph from nodes"""

    # pylint: disable=protected-access
    subonnx = make_graph(nodes, f"subgraph of {graph.name}", inputs=[], outputs=[])
    # inherit initializers and value infos
    subonnx.value_info.extend(graph._value_info)
    initializer_map = {i.name: i for i in graph.initializer}
    for node in nodes:
        for output_name in node.output:
            if output_name in initializer_map:
                initializer_map.pop(output_name)
    for node in nodes:
        for input_name in node.input:
            if input_name in initializer_map:
                subonnx.initializer.append(initializer_map[input_name])
                initializer_map.pop(input_name)

    # append input/output
    for node_name, func_name in inputs:
        shape, dtype = graph.tensor_info(node_name)
        subonnx.input.append(
            make_value_info(func_name, make_tensor_type_proto(dtype, shape))
        )
    for node_name, func_name in outputs:
        shape, dtype = graph.tensor_info(node_name)
        subonnx.output.append(
            make_value_info(func_name, make_tensor_type_proto(dtype, shape))
        )
    sub_model = make_model(
        subonnx,
        doc_string=graph._model.doc_string,
        domain=graph._model.domain,
        model_version=graph._model.model_version,
        ir_version=graph._model.ir_version,
        opset_imports=graph._model.opset_import,
        producer_name=graph._model.producer_name,
        producer_version=graph._model.producer_version,
        functions=_populate_functions(
            {node.op_type for node in nodes}, graph.functions
        ),
    )
    return OnnxGraph(sub_model, base_dir=graph.external_base)



"""
Copyright (C) 2024-2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
from onnx.helper import make_node, np_dtype_to_tensor_dtype
from onnx.onnx_pb import NodeProto, TensorProto

from ...ir.mapping import ETYPE2DTYPE
from .. import OnnxGraph, cast_in
from . import OP_CONVERT, BaseNodeConversion


@OP_CONVERT.register(name="range")
class Range(BaseNodeConversion):
    """https://docs.openvino.ai/2024/documentation/openvino-ir-format/operation-sets/operation-specs/generation/range-4.html

    https://onnx.ai/onnx/operators/onnx__Range.html
    """

    def replace(self, graph: OnnxGraph, ori_node: NodeProto) -> NodeProto:
        output_type = self.get_attribute(ori_node, "output_type")
        if output_type is not None:
            assert isinstance(output_type, str)
            assert output_type in ETYPE2DTYPE, f"Unsupported output_type: {output_type}"
            output_type = ETYPE2DTYPE[output_type]
            output_dtype = np_dtype_to_tensor_dtype(np.dtype(output_type))
        else:
            output_dtype = graph.tensor_type(ori_node.input[0])
        if output_dtype == TensorProto.UNDEFINED:
            output_dtype = TensorProto.INT64
        dtype0 = graph.tensor_type(ori_node.input[0])
        if dtype0 != output_dtype:
            self += cast_in(ori_node, 0, output_dtype)
        dtype1 = graph.tensor_type(ori_node.input[1])
        if dtype1 != output_dtype:
            self += cast_in(ori_node, 1, output_dtype)
        dtype2 = graph.tensor_type(ori_node.input[2])
        if dtype2 != output_dtype:
            self += cast_in(ori_node, 2, output_dtype)
        return make_node(
            "Range",
            inputs=ori_node.input,
            outputs=ori_node.output,
            name=ori_node.name,
        )



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import argparse
from pathlib import Path

import numpy as np

from onnxifier import convert
from onnxifier.evaluator import Evaluator


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("model", type=Path)
    parser.add_argument("-i", "--image", nargs="+")
    parser.add_argument(
        "-a",
        "--activate",
        nargs="*",
        help="select passes to be activated, activate all passes if not set.",
    )
    parser.add_argument(
        "-b1",
        "--backend",
        default="OpenVINO",
        choices=["onnx", "OpenVINO", "OnnxRuntime"],
        help="select a backend to run the original model.",
    )
    parser.add_argument(
        "-b2",
        "--reference-backend",
        default="OnnxRuntime",
        choices=["onnx", "OpenVINO", "OnnxRuntime"],
        help="select a backend to run the converted model.",
    )
    args = parser.parse_args()

    ori_model = Path(args.model).expanduser()
    model = convert(ori_model, args.activate, strict=False)

    ori_runner = Evaluator(str(ori_model), backend=args.backend)
    ort_runner = Evaluator(model, backend=args.reference_backend)

    ori_inputs = {}
    for _, (name, info) in enumerate(ori_runner.inputs.items()):
        try:
            dtype = np.iinfo(info["dtype"]).dtype
            # all zeros for integer types
            ori_inputs[name] = np.zeros(info["shape"], dtype)
        except ValueError:
            ori_inputs[name] = np.random.rand(*info["shape"]).astype(info["dtype"])
    ori_results = ori_runner(sorted(ori_runner.outputs.keys()), ori_inputs)

    ort_inputs = {}
    for name, info in ort_runner.inputs.items():
        if name in ori_inputs:
            ort_inputs[name] = ori_inputs[name]
        else:
            found = False
            for j in ori_inputs.values():
                if tuple(j.shape) == info["shape"]:
                    ort_inputs[name] = j
                    found = True
                    break
            if not found:
                raise ValueError(f"Cannot find input {name} in converted model")
    ort_results = ort_runner(sorted(ort_runner.outputs.keys()), ort_inputs)

    errors = {}
    for name1, name2, ov_ans, ort_ans in zip(
        sorted(ori_runner.outputs), sorted(ort_runner.outputs), ori_results, ort_results
    ):
        if ov_ans.shape != ort_ans.shape:
            raise ValueError(
                "Output shapes do not match: "
                f"{name1}({ov_ans.shape}) vs {name2}({ort_ans.shape})"
            )
        err = np.abs(ov_ans - ort_ans)
        errors[f"{name1}-{name2}"] = (err, ort_ans)
    for k, v in errors.items():
        err, ref = v
        rel = err / (np.abs(ref) + 1e-5)
        print(f"{k}: max error = {np.max(err):.4f}, mean error = {np.mean(err):.4f}")
        print(f"{k}: rel max   = {np.max(rel):.4f}, rel mean   = {np.mean(rel):.4f}")


if __name__ == "__main__":
    main()



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List

import numpy as np
from onnx.helper import make_node
from onnx.onnx_pb import NodeProto

from ...graph import OnnxGraph
from .. import PASSES
from ..pattern import GraphPattern, SingleNodePattern
from ..rewriter import Rewriter
from ..utils import make_constant


@PASSES.register(name="merge_mul_to_conv", deps=["initializer_to_constant"])
class MergeMulToConvRewrite(Rewriter):
    """Merge multiply to the nearest convolution

    Pattern1:

        mul
         |
        add/sub
         |
        conv

    Pattern2:

        conv
         |
        mul

    After:

        conv(bias)
    """

    def __init__(self):
        pattern1 = GraphPattern()
        pattern2 = GraphPattern()
        mul = SingleNodePattern("Mul")
        elt = SingleNodePattern()
        conv = SingleNodePattern("Conv")

        pattern1.add_edge(mul, elt)
        pattern1.add_edge(elt, conv)
        pattern2.add_edge(conv, mul)
        super().__init__(pattern1 | pattern2)

    def _get_input_and_factor(self, mul: NodeProto, graph: OnnxGraph):
        inp0, inp1 = mul.input
        for node in graph:
            pb_node = graph.nodes[node]["pb"]
            # skip not cst op
            if pb_node.op_type != "Constant":
                continue
            # input 0 is cst
            if pb_node.output[0] == inp0:
                return inp1, self.get_value(pb_node)
            # input 1 is cst
            elif pb_node.output[0] == inp1:
                return inp0, self.get_value(pb_node)
        return None, None

    def _make_div(self, input_name: str, op_name: str, factor: np.ndarray):
        cst = make_constant(op_name + "/DivCst", factor)
        div = make_node(
            op_type="Div",
            inputs=[input_name, cst.output[0]],
            outputs=[op_name + "/DivOutput"],
            name=op_name + "/Div",
        )
        return cst, div

    def _make_mul(self, input_name: str, op_name: str, factor: np.ndarray):
        cst = make_constant(op_name + "/MulCst", factor)
        mul = make_node(
            op_type="Mul",
            inputs=[input_name, cst.output[0]],
            outputs=[op_name + "/MulOutput"],
            name=op_name + "/Mul",
        )
        return cst, mul

    def rewrite_pattern1(self, graph: OnnxGraph, nodes: List[NodeProto]):
        """Fuse mul-elt-conv to conv"""
        mul, elt, conv = nodes

        if elt.op_type not in {"Add", "Sub"}:
            return

        # which is constant of mul
        input_name, factor = self._get_input_and_factor(mul, graph)
        if input_name is None or factor is None:
            return

        # which is the input of add/sub op
        in_idx = 0 if mul.output[0] == elt.input[0] else 1

        # change add/sub op input and make new div op
        elt.input[in_idx] = input_name
        cst1, div1 = self._make_div(elt.input[1 - in_idx], elt.name, factor)
        elt.input[1 - in_idx] = div1.output[0]
        self += [cst1, div1]

        # change conv's weight and bias
        cst2, mul2 = self._make_mul(conv.input[1], conv.input[1], factor)
        conv.input[1] = mul2.output[0]
        self += [cst2, mul2]

        self -= [mul]

    def rewrite_pattern2(self, graph: OnnxGraph, nodes: List[NodeProto]):
        """Fuse conv-mul to conv"""
        conv, mul = nodes
        input_name, factor = self._get_input_and_factor(mul, graph)
        if input_name is None or factor is None:
            return
        if len(graph.onnx_successors(conv)) > 1:
            # multi fan-out
            return

        # can not broadcast to weight
        if factor.ndim > 1 and np.prod(factor.shape[-2:]) > 1:
            return

        # change weight
        factor = factor.reshape([-1, 1, 1, 1])
        cst, weight_mul = self._make_mul(conv.input[1], conv.input[1], factor)
        conv.input[1] = weight_mul.output[0]
        # change bias
        if len(conv.input) > 2 and conv.input[2]:
            bias_shape = graph.static_tensor_shape(conv.input[2])
            bias_shape[0] = -1
            factor = factor.reshape(bias_shape)
            cst2, bias_mul = self._make_mul(conv.input[2], conv.input[2], factor)
            self += [cst2, bias_mul]
            conv.input[2] = bias_mul.output[0]
        # for post_mul in graph.onnx_successors(mul):
        #     post_mul.input[list(post_mul.input).index(mul.output[0])] = conv.output[0]
        conv.output[0] = mul.output[0]
        self += [cst, weight_mul, conv]
        self -= mul

    # pylint: disable=arguments-differ
    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto]):
        if len(nodes) == 3:
            self.rewrite_pattern1(graph, nodes)
        elif len(nodes) == 2:
            self.rewrite_pattern2(graph, nodes)



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
from onnx import TensorProto
from onnx.helper import (
    make_graph,
    make_model,
    make_node,
    make_tensor,
    make_tensor_type_proto,
    make_value_info,
)

from onnxifier.graph import OnnxGraph
from onnxifier.passes.canonicalization.resize_move_size_to_scale import (
    ResizeMoveSizeToScaleRewriter,
)
from onnxifier.passes.swap.initializer_to_constant import initializer_to_constant


def _build_test_graph():
    resize = make_node(
        "Resize",
        inputs=["x", "roi", "scales", "sizes"],
        outputs=["y"],
        name="resize",
        coordinate_transformation_mode="tf_crop_and_resize",
        mode="cubic",
        axes=[2, 3],
    )
    graph = make_graph(
        [resize],
        "graph",
        [make_value_info("x", make_tensor_type_proto(1, [1, 3, 8, 8]))],
        [make_value_info("y", make_tensor_type_proto(1, None))],
        [
            make_tensor(
                "roi", TensorProto.INT64, [4], np.array([2, 2, 6, -2], "int64")
            ),
            make_tensor("sizes", TensorProto.INT64, [2], np.array([8, 8], "int64")),
        ],
    )
    return make_model(graph)


def test_roi_rewriter():
    graph = OnnxGraph(_build_test_graph())
    rewriter = ResizeMoveSizeToScaleRewriter()
    graph = rewriter(initializer_to_constant(graph))



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Dict, List

from ... import OnnxGraph
from .. import PASSES


@PASSES.register("reshape_model", deps=["infer_shape"])
def reshape_model(graph: OnnxGraph, shape_info: Dict[str, List[int | str]]):
    """Update model input and output dimensions.

    Args:
        shape_info (Dict[str, List[int | str]]): A dictionary of I/O names and
            their new shapes.

    Example:

        onnxify model.onnx -a reshape_model --config-file config.json

        # config.json:
        {
            "reshape_model:0": {
                "shape_info": {
                    "input": [1, 2, 3, 4],
                    "output": [4, 3, 2, 1]
                }
            }
        }
    """

    def _update_dim(dims, shape):
        for d, i in zip(dims, shape):
            if isinstance(i, int):
                if i < 0 and i != -1:
                    raise ValueError(f"negative value in shape must be -1, but got {i}")
                d.dim_value = i
            elif isinstance(i, str):
                d.dim_param = i
            else:
                raise TypeError(f"shape must be int or str, but got {type(i)}")

    for _, input_info in enumerate(graph.input):
        shape = shape_info[input_info.name]
        _update_dim(input_info.type.tensor_type.shape.dim, shape)
    for _, output_info in enumerate(graph.output):
        shape = shape_info[output_info.name]
        _update_dim(output_info.type.tensor_type.shape.dim, shape)
    return graph


@PASSES.register("reset_model_batch", deps=["infer_shape"])
def reset_model_batch(graph: OnnxGraph, **kwargs):
    """Reset the batch size of the model.

    Batch size is the 1st dimension of all tensors.

    Note:

        This pass does not check the feasibility of the new batch size. And
        it will fail to infer shapes if the new batch size is not compatible
        with the previous shapes.
    """
    kwargs.setdefault("batch", 1)
    kwargs.setdefault("batch_size", 1)
    batch = kwargs.get("batch") or kwargs.get("batch_size")
    assert isinstance(batch, int) and batch > 0
    for _, input_info in enumerate(graph.input):
        input_info.type.tensor_type.shape.dim[0].dim_value = batch
    for _, output_info in enumerate(graph.output):
        output_info.type.tensor_type.shape.dim[0].dim_value = batch
    return graph



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from collections.abc import Sequence
from typing import List

import numpy as np
from onnx.onnx_pb import NodeProto

from ... import OnnxGraph
from .. import PASSES
from ..pattern import GraphPattern, SingleNodePattern
from ..rewriter import Rewriter
from ..utils import make_constant


@PASSES.register(name="swap_concat_input_order")
class SwapConcatInputOrderRewriter(Rewriter):
    """Swap the order of concat inputs.

    This rewriter is only acceptable if concat axis is 1.

    Before:

        concat(a, b, c)

    After:

        concat(c, b, a)  # [2, 1, 0]
    """

    def __init__(self):
        pattern = GraphPattern()
        concat = SingleNodePattern("Concat").with_attr("axis", 1)
        conv = SingleNodePattern("Conv")
        pattern.add_edge(concat, conv)
        super().__init__(pattern)

    # pylint: disable=arguments-differ
    def rewrite(
        self,
        graph: OnnxGraph,
        nodes: List[NodeProto],
        order: Sequence[int] | None = None,
    ):
        if not order:
            raise ValueError("swap order is not specified, skip rewriter")
        order = list(order)
        if len(order) != len(nodes[0].input):
            raise ValueError(
                f"swap order {order} must be same as concat input number "
                f"{len(nodes[0].input)}."
            )

        concat, conv = nodes
        if self.get_attribute(conv, "group", 1) != 1:
            raise ValueError("Conv group is not 1, skip rewriter")

        weight_data = self.get_value_or_die(conv.input[1])
        channels = [graph.static_tensor_shape(i)[1] for i in concat.input]
        new_inputs = np.asarray(concat.input)[order].tolist()
        concat.input[:] = new_inputs
        indices = np.cumsum(channels)[:-1]

        weight_sections = np.split(weight_data, indices, axis=1)
        new_weight = np.concatenate([weight_sections[i] for i in order], axis=1)
        weight_node = make_constant(f"{conv.name}/weight_shuffled", new_weight)
        conv.input[1] = weight_node.output[0]
        self += weight_node



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
from onnx import TensorProto
from onnx.helper import (
    make_graph,
    make_model,
    make_node,
    make_tensor_type_proto,
    make_value_info,
)
from onnx.numpy_helper import from_array

from onnxifier import ONNXIFIER_IR_VERSION, ONNXIFIER_OPSET, PassManager
from onnxifier.graph import OnnxGraph
from onnxifier.passes.utils import evaluate_on_node


def _make_scatter_nd_model(indices_data, data_shape):
    """Create a ScatterND model with given indices and data shape."""
    # Create input data tensor info
    data_info = make_value_info(
        "data", make_tensor_type_proto(TensorProto.FLOAT, data_shape)
    )

    # For ScatterND, updates shape is
    # indices.shape[:-1] + data_shape[indices.shape[-1]:]
    updates_shape = list(indices_data.shape[:-1]) + data_shape[indices_data.shape[-1] :]
    updates_info = make_value_info(
        "updates", make_tensor_type_proto(TensorProto.FLOAT, updates_shape)
    )

    # Create output tensor info
    output_info = make_value_info(
        "output", make_tensor_type_proto(TensorProto.FLOAT, data_shape)
    )

    # Create indices initializer
    indices_tensor = from_array(indices_data, "indices")

    # Create scatter node
    scatter_node = make_node(
        "ScatterND", ["data", "indices", "updates"], ["output"], name="scatter_nd"
    )

    # Create graph
    graph = make_graph(
        [scatter_node],
        "scatter_nd_graph",
        [data_info, updates_info],
        [output_info],
        [indices_tensor],
    )

    # Create model
    model = make_model(
        graph, ir_version=ONNXIFIER_IR_VERSION, opset_imports=[ONNXIFIER_OPSET]
    )
    return model


def test_no_negative_indices():
    """Test case where indices are already positive."""
    # Create indices with no negative values
    # For data_shape [2, 3, 4], indices should have shape [..., 3]
    # where 3 is the rank of data
    indices = np.array([[0, 1, 2], [1, 2, 3]], dtype=np.int64)
    data_shape = [2, 3, 4]

    model = _make_scatter_nd_model(indices, data_shape)
    graph = OnnxGraph(model)

    # Apply pass
    pm = PassManager(["scatter_nd_indices_to_positive"])
    graph = pm.optimize(graph, strict=True)

    assert "scatter_nd/indices_pos" not in graph


def test_with_negative_indices():
    """Test case with negative indices that should be converted to positive."""
    indices = np.array([[0, 1, -1], [1, -1, 3]], dtype=np.int64)
    data_shape = [2, 3, 4]

    model = _make_scatter_nd_model(indices, data_shape)
    graph = OnnxGraph(model)

    # Apply pass
    pm = PassManager(["scatter_nd_indices_to_positive"])
    graph = pm.optimize(graph, strict=True)

    rewritten_indices = evaluate_on_node(
        graph, graph.nodes["scatter_nd/indices_pos"]["pb"]
    )

    # Expected result: [[0, 1, 3], [1, 2, 3]]
    # (since -1 in dim 2 becomes 3, and -1 in dim 1 becomes 2)
    assert rewritten_indices is not None
    expected_indices = np.array([[0, 1, 3], [1, 2, 3]], dtype=np.int64)
    assert np.array_equal(rewritten_indices, expected_indices)



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import onnx
from onnx.helper import make_graph, make_node, make_tensor_value_info
from onnx.reference import ReferenceEvaluator

from onnxifier import OnnxGraph, PassManager


def _build_graph():
    resize = make_node("Resize", ["x", "", "", "sizes"], ["y"])
    graph = make_graph(
        [resize],
        "test",
        [make_tensor_value_info("x", onnx.TensorProto.FLOAT, [4, 3, 4, 4])],
        [make_tensor_value_info("y", onnx.TensorProto.FLOAT, [4, 3, 8, 8])],
        initializer=[
            onnx.numpy_helper.from_array(
                np.array([4, 3, 8, 8], dtype="int64"), "sizes"
            ),
        ],
    )
    model = onnx.helper.make_model(graph)
    onnx.checker.check_model(model, full_check=True)
    return model


def test_change_resize_batch():
    graph = OnnxGraph(_build_graph())
    runner1 = ReferenceEvaluator(graph.model)
    x = np.random.uniform(0, 1, size=[4, 3, 4, 4]).astype(np.float32)
    y1 = runner1.run(None, {"x": x})[0]

    pm = PassManager(
        ["initlializer_to_constant", "resize_move_size_to_scale", "change_resize_batch"]
    )
    graph = pm.optimize(graph, strict=True)
    runner2 = ReferenceEvaluator(graph.model)
    y2 = runner2.run(None, {"x": x})
    assert np.allclose(y1, y2)



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Choose a proper backend to evaluate ONNX models.
1. ReferenceEvaluator: no extra dependency, but extremely slow.
2. OpenVINO runtime: best for Intel platforms.
3. OnnxRuntime: best compatibility and balanced performance.
"""

# pylint: disable=import-outside-toplevel

import os
import warnings
from contextlib import suppress
from tempfile import TemporaryDirectory
from typing import Any, Callable, Dict, List, Literal, Sequence, Tuple

import numpy as np
import onnx
from onnx.helper import tensor_dtype_to_np_dtype


def _get_eval_onnx(model):
    from onnx.reference import ReferenceEvaluator

    if not isinstance(model, onnx.ModelProto):
        model = onnx.load_model(model)

    inputs = {}
    outputs = {}
    for i in model.graph.input:
        shape = [int(j.dim_value) for j in i.type.tensor_type.shape.dim]
        inputs[i.name] = dict(
            shape=shape,
            dtype=tensor_dtype_to_np_dtype(i.type.tensor_type.elem_type),
        )
    for i in model.graph.output:
        shape = [int(j.dim_value) for j in i.type.tensor_type.shape.dim]
        outputs[i.name] = dict(
            shape=shape,
            dtype=tensor_dtype_to_np_dtype(i.type.tensor_type.elem_type),
        )

    model = ReferenceEvaluator(model)

    def _run_code(output_names, inputs_feed):
        results = model.run(output_names, inputs_feed)
        return [np.array(i) for i in results]

    return _run_code, inputs, outputs


def _get_eval_onnxruntime(model):
    import onnxruntime

    if isinstance(model, onnx.ModelProto):
        sess = onnxruntime.InferenceSession(model.SerializeToString())
    else:
        sess = onnxruntime.InferenceSession(model)

    def _run_code(
        output_names: Sequence[str] | None, inputs_feed: Dict[str, np.ndarray]
    ) -> List[Any]:
        return list(sess.run(output_names, inputs_feed))

    def _data_mapping():
        mapping = {}
        for dtype in (
            np.float64,
            np.float32,
            np.half,
            np.uint64,
            np.uint32,
            np.uint16,
            np.uint8,
            np.int64,
            np.int32,
            np.int16,
            np.int8,
            np.bool_,
        ):
            mapping[
                onnxruntime.OrtValue.ortvalue_from_numpy(np.ones([], dtype)).data_type()
            ] = dtype
        return mapping

    inputs = {}
    outputs = {}
    mapping = _data_mapping()
    for i in sess.get_inputs():
        inputs[i.name] = dict(shape=tuple(i.shape), dtype=mapping[i.type])
    for i in sess.get_outputs():
        outputs[i.name] = dict(shape=tuple(i.shape), dtype=mapping[i.type])

    return _run_code, inputs, outputs


def _get_eval_openvino(model):
    import openvino

    core = openvino.Core()
    with TemporaryDirectory() as tmpdir:
        if isinstance(model, onnx.ModelProto):
            onnx.save_model(model, f"{tmpdir}/model.onnx")
            network = core.read_model(f"{tmpdir}/model.onnx")
            model = openvino.compile_model(network)
        else:
            network = core.read_model(model)
            model = core.compile_model(network)

    def _run_code(output_names, inputs_feed):
        input_names = {i.name for i in network.get_parameters()}
        if missing_names := input_names.difference(inputs_feed.keys()):
            warnings.warn(f"Missing inputs: {missing_names}")
        outputs = model(inputs_feed)
        if output_names:
            return [outputs[name] for name in output_names]
        return [outputs[i] for i, _ in enumerate(outputs)]

    inputs = {}
    outputs = {}
    for i, inp in enumerate(network.inputs):
        if inp.names:
            key = inp.get_any_name()
        else:
            key = i
        inputs[key] = dict(
            shape=tuple(inp.shape), dtype=inp.get_element_type().to_dtype()
        )
    for i, oup in enumerate(network.outputs):
        if oup.names:
            key = oup.get_any_name()
        else:
            key = i
        outputs[key] = dict(
            shape=tuple(oup.shape), dtype=oup.get_element_type().to_dtype()
        )

    return _run_code, inputs, outputs


# pyright: ignore[reportReturnType]
def _get_eval_backend(
    backend: str, model: str | os.PathLike | onnx.ModelProto
) -> Tuple[
    Callable[[Sequence[str] | None, Dict[str, np.ndarray]], List[np.ndarray]],
    Dict,
    Dict,
]:
    if backend == "onnx":
        return _get_eval_onnx(model)
    elif backend == "onnxruntime":
        return _get_eval_onnxruntime(model)
    elif backend == "openvino":
        return _get_eval_openvino(model)

    raise NotImplementedError(f"Unsupported backend: {backend}")


class Evaluator:
    """An evaluator wraps different backends to evaluate ONNX models."""

    def __init__(
        self,
        model: str | os.PathLike | onnx.ModelProto,
        backend: Literal["auto", "openvino", "onnxruntime", "onnx"] = "auto",
    ):
        if backend.lower() == "auto":
            with suppress(ImportError):
                # pylint: disable=unused-import
                import onnxruntime  # noqa: F401

                backend = "onnxruntime"
        if backend.lower() == "auto":
            with suppress(ImportError):
                # pylint: disable=unused-import
                import openvino  # noqa: F401

                backend = "openvino"
        if backend.lower() == "auto":
            backend = "onnx"

        self._call_fn, inputs, outputs = _get_eval_backend(backend.lower(), model)
        self.inputs = inputs
        self.outputs = outputs

    def __call__(
        self,
        output_names: Sequence[str] | None,
        feed_inputs: Dict[str, np.ndarray] | None = None,
        **kw_feeds: np.ndarray,
    ) -> List[np.ndarray]:
        if feed_inputs is None and not kw_feeds:
            raise ValueError(
                "Either feed_inputs or **kw_feeds should be provided."
                " But both are None."
            )
        model_inputs: Dict[str, np.ndarray] = {}
        if feed_inputs:
            model_inputs.update(feed_inputs)
        if kw_feeds:
            model_inputs.update(kw_feeds)
        return self._call_fn(output_names, model_inputs)



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
from onnx.helper import (
    make_graph,
    make_model,
    make_node,
    make_tensor,
    make_tensor_type_proto,
    make_value_info,
)
from onnx.reference import ReferenceEvaluator

from onnxifier import PassManager
from onnxifier.graph import OnnxGraph


def _build_graph():
    dq1 = make_node("DequantizeLinear", ["x1", "s1", "zp1"], ["y1"], "dq1", axis=0)
    dq2 = make_node("DequantizeLinear", ["x2", "s2", "zp2"], ["y2"], "dq2", axis=0)
    concat = make_node("Concat", ["y1", "y2"], ["z"], "concat", axis=0)
    graph = make_graph(
        [dq1, dq2, concat],
        "graph",
        [
            make_value_info("x1", make_tensor_type_proto(1, [1, 16, 4, 4])),
            make_value_info("x2", make_tensor_type_proto(1, [1, 16, 4, 4])),
        ],
        [make_value_info("z", make_tensor_type_proto(1, [2, 16, 4, 4]))],
        [
            make_tensor("s1", 1, [1], np.random.rand(1).astype("float32")),
            make_tensor("s2", 1, [1], np.random.rand(1).astype("float32")),
            make_tensor(
                "zp1", 3, [1], np.random.randint(0, 127, size=[1], dtype="int8")
            ),
            make_tensor(
                "zp2", 3, [1], np.random.randint(0, 127, size=[1], dtype="int8")
            ),
        ],
    )
    return make_model(graph)


def _build_graph2():
    dq1 = make_node("DequantizeLinear", ["x1", "s1", "zp1"], ["y1"], "dq1", axis=0)
    dq2 = make_node("DequantizeLinear", ["x2", "s2", "zp2"], ["y2"], "dq2", axis=0)
    dq3 = make_node("DequantizeLinear", ["x3", "s3", "zp3"], ["y3"], "dq3", axis=0)
    concat = make_node("Concat", ["y1", "y2", "y3"], ["z"], "concat", axis=0)
    graph = make_graph(
        [dq1, dq2, dq3, concat],
        "graph",
        [
            make_value_info("x1", make_tensor_type_proto(1, [1, 16, 4, 4])),
            make_value_info("x2", make_tensor_type_proto(1, [1, 16, 4, 4])),
            make_value_info("x3", make_tensor_type_proto(1, [1, 16, 4, 4])),
        ],
        [make_value_info("z", make_tensor_type_proto(1, [2, 16, 4, 4]))],
        [
            make_tensor("s1", 1, [1], np.random.rand(1).astype("float32")),
            make_tensor("s2", 1, [1], np.random.rand(1).astype("float32")),
            make_tensor("s3", 1, [1], np.random.rand(1).astype("float32")),
            make_tensor(
                "zp1", 3, [1], np.random.randint(0, 127, size=[1], dtype="int8")
            ),
            make_tensor(
                "zp2", 3, [1], np.random.randint(0, 127, size=[1], dtype="int8")
            ),
            make_tensor(
                "zp3", 3, [1], np.random.randint(0, 127, size=[1], dtype="int8")
            ),
        ],
    )
    return make_model(graph)


def test_swap_concat():
    model = _build_graph()
    runner1 = ReferenceEvaluator(model)
    x1 = np.random.randint(-128, 127, size=[1, 16, 4, 4], dtype="int8")
    x2 = np.random.randint(-128, 127, size=[1, 16, 4, 4], dtype="int8")
    y1 = runner1.run(None, {"x1": x1, "x2": x2})[0]

    graph = OnnxGraph(model)
    pm = PassManager(["swap_concat"])
    graph = pm.optimize(graph, strict=True)

    # check output
    runner2 = ReferenceEvaluator(graph.model)
    y2 = runner2.run(None, {"x1": x1, "x2": x2})[0]
    assert np.allclose(y1, y2)

    # check conv number
    conv_number = 0
    for name in graph:
        if graph.nodes[name]["pb"].op_type == "DequantizeLinear":
            conv_number += 1
    assert conv_number == 1


def test_swap_concat2():
    model = _build_graph2()
    runner1 = ReferenceEvaluator(model)
    x1 = np.random.randint(-128, 127, size=[1, 16, 4, 4], dtype="int8")
    x2 = np.random.randint(-128, 127, size=[1, 16, 4, 4], dtype="int8")
    x3 = np.random.randint(-128, 127, size=[1, 16, 4, 4], dtype="int8")
    y1 = runner1.run(None, {"x1": x1, "x2": x2, "x3": x3})[0]

    graph = OnnxGraph(model)
    pm = PassManager(["swap_concat"])
    graph = pm.optimize(graph, strict=True)

    # check output
    runner2 = ReferenceEvaluator(graph.model)
    y2 = runner2.run(None, {"x1": x1, "x2": x2, "x3": x3})[0]
    assert np.allclose(y1, y2)

    # check conv number
    conv_number = 0
    for name in graph:
        if graph.nodes[name]["pb"].op_type == "DequantizeLinear":
            conv_number += 1
    assert conv_number == 1



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from contextlib import suppress
from typing import List, Optional

from ... import OnnxGraph
from .. import L3

with suppress(ImportError):
    import onnxoptimizer

    @L3.register()
    def onnx_optimizer(graph: OnnxGraph, passes: Optional[List[str]] = None):
        """Fuse op and remove isolated nodes.

        Args:
            graph (OnnxGraph): The onnx graph to be optimized.
            passes (List[str], optional): The optimization passes to be applied.
                Defaults to None.
        """

        if not passes:
            passes = [
                "eliminate_nop_cast",
                "eliminate_nop_dropout",
                "eliminate_nop_flatten",
                "eliminate_if_with_const_cond",
                "eliminate_nop_monotone_argmax",
                "eliminate_nop_pad",
                "eliminate_nop_concat",
                "eliminate_nop_split",
                "eliminate_nop_expand",
                "eliminate_shape_gather",
                "eliminate_slice_after_shape",
                "eliminate_nop_reshape",
                "eliminate_nop_with_unit",
                "eliminate_common_subexpression",
                "eliminate_deadend",
                "eliminate_identity",
                "eliminate_shape_op",
                "eliminate_unused_initializer",
                "eliminate_duplicate_initializer",
                "fuse_add_bias_into_conv",
                "fuse_bn_into_conv",
                "fuse_concat_into_reshape",
                "fuse_consecutive_concats",
                "fuse_consecutive_log_softmax",
                "fuse_consecutive_reduce_unsqueeze",
                "fuse_consecutive_slices",
                "fuse_consecutive_squeezes",
                "fuse_consecutive_transposes",
                "fuse_consecutive_unsqueezes",
                "fuse_matmul_add_bias_into_gemm",
                "fuse_pad_into_conv",
                "fuse_pad_into_pool",
                "fuse_qkv",
                "fuse_transpose_into_gemm",
                "extract_constant_to_initializer",
            ]
        # FIXME: onnxoptimizer ignores functions
        opt_model = onnxoptimizer.optimize(graph.model, passes)
        opt_model.functions.extend(graph.functions.values())
        return OnnxGraph(opt_model, base_dir=graph.external_base)



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import onnx
import pytest
from onnx.helper import make_graph, make_model, make_node, make_tensor_value_info
from onnx.numpy_helper import from_array

from onnxifier import ONNXIFIER_IR_VERSION, ONNXIFIER_OPSET, OnnxGraph, PassManager
from onnxifier.evaluator import Evaluator


def _build_graph():
    resize = make_node(
        "Resize", ["x", "", "scales"], ["y"], axes=[2, 3, 4], name="resize"
    )
    graph = make_graph(
        [resize],
        "test",
        [make_tensor_value_info("x", onnx.TensorProto.FLOAT, [1, 3, 64, 128, 256])],
        [make_tensor_value_info("y", onnx.TensorProto.FLOAT, [1, 3, 32, 32, 32])],
        [from_array(np.array([0.5, 0.25, 0.125], np.float32), "scales")],
    )
    model = make_model(
        graph,
        ir_version=ONNXIFIER_IR_VERSION,
        opset_imports=[ONNXIFIER_OPSET],
    )
    return OnnxGraph(model)


def test_resize_remove_axes():
    graph = _build_graph()
    try:
        run1 = Evaluator(graph.model, "OnnxRuntime")
        x = np.random.randn(1, 3, 64, 128, 256).astype(np.float32)
        y1 = run1(["y"], {"x": x})[0]
        assert y1.shape == (1, 3, 32, 32, 32)

        pm = PassManager(["resize_remove_axes"])
        graph = pm.optimize(graph, strict=True)
        run2 = Evaluator(graph.model, "OnnxRuntime")
        y2 = run2(["y"], {"x": x})[0]
        assert np.allclose(y1, y2)
    except ImportError:
        pytest.skip("onnxruntime not installed")



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from copy import deepcopy
from itertools import chain
from typing import List

import networkx as nx
from onnx import NodeProto

from ... import OnnxGraph
from ...logger import debug
from .. import PASSES
from ..pattern import SingleNodePattern, StartEndPointPattern
from ..rewriter import Rewriter
from ..utils import is_elewise, make_constant


@PASSES.register("eliminate_slice_after_conv", deps=["split_to_slice"])
class EliminateSliceAfterConvRewriter(Rewriter):
    r"""Eliminate slice(split) after conv by splitting output channels of conv
    into multiple pieces.

    Before:

        conv -> * -> split -> A
                          \__ B

    After:

        conv_1 -> * -> A
        conv_2 -> * -> B

    Note:
        The split must be taken on channel axis.
    """

    def __init__(self):
        # TODO: add ConvTranspose support
        super().__init__(pattern=SingleNodePattern("Conv"))

    def is_slice_on_channel(self, node: NodeProto, graph: OnnxGraph) -> bool:
        """Whether the slice is on channel axis only (axis=1).

        output = Slice(input, starts, ends, axes, steps)
        """

        assert node.op_type == "Slice"
        if len(node.input) >= 4:
            axes = self.get_value(node.input[3])
            if axes is None:
                # dynamic axes
                return False
            if len(axes) == 1:
                # whether on channel axis
                return axes[0] == 1
        else:
            axes = None
        data_shape, _ = graph.tensor_info(node.input[0])
        if data_shape is None:
            # dynamic data shape
            return False
        if axes is None:
            axes = list(range(len(data_shape)))
        if len(axes) > 1 and 1 in axes:
            starts = self.get_value(node.input[1])
            ends = self.get_value(node.input[2])
            if starts is None or ends is None:
                # dynamic slice
                return False
            data_shape, _ = graph.tensor_info(node.input[0])
            if data_shape is None:
                # dynamic data shape
                return False
            steps = (
                self.get_value(node.input[4])
                if len(node.input) >= 5
                else [1 for _ in axes]
            )
            if steps is None:
                # dynamic steps
                return False
            for start, end, axis, step in zip(starts, ends, axes, steps):
                if axis == 1:
                    continue
                if start != 0 or end != data_shape[axis] or step != 1:
                    # not on channel axis
                    return False
            return True
        return False

    def get_sliced_axes(self, node: NodeProto, graph: OnnxGraph) -> slice:
        """Get the axes that are sliced by the slice node."""
        assert node.op_type == "Slice"
        data_shape = graph.tensor_shape(node.input[0])
        axes = self.get_value(node.input[3]) if len(node.input) >= 4 else None
        if axes is None:
            axes = list(range(len(data_shape)))
        steps = (
            self.get_value(node.input[4]) if len(node.input) >= 5 else [1 for _ in axes]
        )
        starts = self.get_value(node.input[1])
        ends = self.get_value(node.input[2])
        assert starts is not None and ends is not None and steps is not None

        for start, end, axis, step in zip(starts, ends, axes, steps):
            if axis != 1:
                continue
            return slice(start, end, step)

        raise RuntimeError("Slice is not on channel axis.")

    def make_split_conv(self, conv: NodeProto, axes: slice, graph: OnnxGraph):
        """Split the output channels of conv into multiple pieces."""
        # TODO: support quantized conv
        weights = self.get_value(conv.input[1])
        if weights is None:
            raise RuntimeError("Quantized conv or dynamic weight is not supported.")
        bias = self.get_value(conv.input[2]) if len(conv.input) >= 3 else None
        new_weights = weights[axes]
        new_bias = bias[axes] if bias is not None else None
        node_conv = deepcopy(conv)
        node_conv.output[0] = f"{conv.name}/{axes.start}_{axes.stop}_{axes.step}_output"
        node_conv.name = f"{conv.name}/{axes.start}_{axes.stop}_{axes.step}"
        node_weight = make_constant(f"{node_conv.name}/weight", new_weights)
        node_conv.input[1] = node_weight.output[0]
        self += node_weight
        if new_bias is not None:
            node_bias = make_constant(f"{node_conv.name}/bias", new_bias)
            node_conv.input[2] = node_bias.output[0]
            self += node_bias
        self += node_conv
        return node_conv

    def copy_relay_nodes(
        self, conv: NodeProto, nodes: List[NodeProto], graph: OnnxGraph
    ):
        """Duplicate the relay nodes between conv (start) and slice (end) node."""
        # TODO: support sibling nodes.
        # E.g. Add(a, b) where a is split and b is channel-wise constant.
        h = graph.onnx_subgraph(nodes)
        start, end, *relay_nodes = nodes
        new_relay_nodes = {i.name: i for i in deepcopy(relay_nodes)}
        for node in new_relay_nodes.values():
            node.name += f"/{conv.name}"
            for i, _ in enumerate(node.input):
                node.input[i] += f"/{conv.name}"
            for i, _ in enumerate(node.output):
                node.output[i] += f"/{conv.name}"
        for node in h.onnx_successors(start):
            new_node = new_relay_nodes[node.name]
            for i, input_name in enumerate(node.input):
                if input_name == start.output[0]:
                    new_node.input[i] = conv.output[0]
        for node in h.onnx_predecessors(end):
            new_node = new_relay_nodes[node.name]
            for i, output_name in enumerate(node.output):
                if output_name == end.input[0]:
                    new_node.output[i] = end.output[0]
        self += list(new_relay_nodes.values())

    def rewrite(self, graph, nodes, *args, **kwargs):
        conv = nodes[0]
        search_results = self.search_slices(conv, graph)
        for _, post_nodes in enumerate(search_results):
            slice_nodes = []
            relay_nodes = []
            post_nodes: List[NodeProto] = list(post_nodes)
            for relay_node in post_nodes:
                if relay_node.name == conv.name:
                    continue
                if not is_elewise(relay_node):
                    if relay_node.op_type == "Slice":
                        slice_nodes.append(relay_node)
                    else:
                        slice_nodes.clear()  # invalid branch found, skip
                        break
                else:
                    relay_nodes.append(relay_node)
            if len(slice_nodes) == 1:
                if not self.is_slice_on_channel(slice_nodes[0], graph):
                    continue
            else:
                continue
            axes = self.get_sliced_axes(slice_nodes[0], graph)
            new_conv = self.make_split_conv(conv, axes, graph)
            debug(f"relay nodes: {[i.name for i in relay_nodes]}")
            self.copy_relay_nodes(new_conv, [conv, slice_nodes[0], *relay_nodes], graph)
            self -= slice_nodes

    def search_slices(self, conv: NodeProto, graph: OnnxGraph):
        """Search for slice nodes after conv."""
        h = graph.onnx_subgraph(chain(nx.descendants(graph, conv.name), [conv]))
        # limit search depth to 10
        subgraph_nodes = [conv.name]
        for child in h:
            if nx.shortest_path_length(h, conv.name, child) <= 10:
                subgraph_nodes.append(child)
        # renew h
        h = graph.onnx_subgraph(subgraph_nodes)
        search_results = StartEndPointPattern(
            start_pattern=SingleNodePattern(conv.op_type, conv.name),
            end_pattern=SingleNodePattern("Slice"),
        ).match(h)
        yield from search_results



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List

from onnx.onnx_pb import NodeProto

from ... import OnnxGraph
from .. import PASSES
from ..pattern import SingleNodePattern
from ..rewriter import Rewriter


@PASSES.register(name="canonicalize_conv_autopad")
class CanonicalizeConvAutoPadRewriter(Rewriter):
    """After onnxsim it may combine pad into Conv, but still keep auto_pad attribute.
    This will cause an error in onnxruntime. So we need to remove the auto_pad
    attribute if pads is also set.
    """

    def __init__(self):
        super().__init__(
            SingleNodePattern("Conv").with_attr("auto_pad").with_attr("pads")
        )

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto], *args, **kwargs):
        node = nodes[0]
        auto_pad = self.get_attribute(node, "auto_pad")
        if auto_pad != "NOTSET":
            self.set_attribute(node, "auto_pad", "NOTSET")



"""
Copyright (C) 2024-2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List, Literal, Optional

import numpy as np
from onnx.onnx_pb import NodeProto

from .... import OnnxGraph
from ...pattern import SingleNodePattern
from ...rewriter import Rewriter
from ...utils import make_constant
from . import OP_CONVERTER


@OP_CONVERTER.register("Resize", deps=["resize_move_size_to_scale"])
class Resize(Rewriter):
    """Remove new attributes in Resize operator.

    - antialias - INT (default 0)
    - axes - INTS
    - keep_aspect_ratio_policy - STRING (default "stretch")
    """

    def __init__(self):
        super().__init__(SingleNodePattern("Resize"))

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto], *args, **kwargs):
        node = nodes[0]
        mode = self.get_attribute(node, "mode")
        antialias = self.get_attribute(node, "antialias")
        if antialias and mode in ("linear", "cubic"):
            raise ValueError(
                f"Resize({mode}) with antialiasing not supported below opset 18."
            )
        self.remove_attribute(node, "antialias")
        if ori_axes := self.get_attribute(node, "axes"):
            input_shape = graph.tensor_shape(node.input[0])
            input_rank = len(input_shape)
            axes = [i if i >= 0 else i + input_rank for i in ori_axes]  # type: ignore
            missing_axes = set(range(input_rank)) - set(axes)
            scales_node = self.get_input_node_or_die(node, 2)
            scales = self.get_value_or_die(scales_node)
            dtype = scales.dtype
            scales_value = {axis: scales[i] for i, axis in enumerate(axes)}
            scales_value.update({axis: 1.0 for axis in missing_axes})
            scales_value = [scales_value[axis] for axis in sorted(scales_value)]
            scales = np.array(scales_value, dtype=dtype)
            new_scales = make_constant(scales_node.name + "/new", scales)
            node.input[2] = new_scales.output[0]
            self -= scales_node
            self += new_scales
            keep_aspect_ratio_policy = self.get_attribute(
                node, "keep_aspect_ratio_policy"
            )
            keep_aspect_ratio_policy = f"{keep_aspect_ratio_policy}"  # force __repr__
            if keep_aspect_ratio_policy in ("not_larger", "not_smaller"):
                # rescale should use the original axes to specify the min/max range of
                # scales value.
                self._rescale(
                    node,
                    keep_aspect_ratio_policy,
                    axes,
                    scales,
                )
        self.remove_attribute(node, "axes")
        self.remove_attribute(node, "keep_aspect_ratio_policy")

    def _rescale(
        self,
        node: NodeProto,
        policy: Literal["not_larger", "not_smaller"],
        axes: Optional[List[int]],
        scales: np.ndarray,
    ):
        if axes is None:
            axes = list(range(len(scales)))
        axes = sorted(axes)
        scales = scales[axes]
        if policy == "not_larger":
            scales = scales.min()
        else:
            scales = scales.max()
        scales = np.tile(scales, [len(axes)])
        # append 1.0 for missing axes
        input_shape = self.graph.tensor_shape(node.input[0])
        new_scales = np.ones([len(input_shape)], dtype=scales.dtype)
        scales_iter = iter(scales)
        for i, _ in enumerate(input_shape):
            new_scales[i] = next(scales_iter) if i in axes else 1.0
        scales_cst = make_constant(f"{node.name}/new_scales", new_scales)
        node.input[2] = scales_cst.output[0]
        self += scales_cst



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
from onnx.helper import make_node, tensor_dtype_to_np_dtype
from onnx.onnx_pb import NodeProto

from .. import OnnxGraph, make_constant
from . import OP_CONVERT, BaseNodeConversion


@OP_CONVERT.register(name="fake_quantize")
class FakeQuantize(BaseNodeConversion):
    """https://docs.openvino.ai/2024/documentation/openvino-ir-format/operation-sets/operation-specs/quantization/fake-quantize-1.html

    Split fake quantize to QuantizeLinear and DequantizeLinear.
    https://onnx.ai/onnx/operators/onnx__QuantizeLinear.html
    https://onnx.ai/onnx/operators/onnx__DequantizeLinear.html

    Note:

        FakeQuantize only applies to activations, not weights.
    """

    def replace(self, graph: OnnxGraph, ori_node: NodeProto) -> NodeProto:
        levels = int(self.get_attribute(ori_node, "levels") or 256)  # type: ignore
        nbits = int(np.log2(levels))
        assert nbits in (8, 16, 32)
        input_low = self.get_value_or_die(ori_node.input[1]).squeeze()
        input_high = self.get_value_or_die(ori_node.input[2]).squeeze()
        output_low = self.get_value_or_die(ori_node.input[3]).squeeze()
        output_high = self.get_value_or_die(ori_node.input[4]).squeeze()

        try:
            scale_type = graph.tensor_type(ori_node.output[0])
            scale_prec = tensor_dtype_to_np_dtype(scale_type)
        except ValueError:
            scale_prec = np.float32
        zero_prec = np.iinfo(f"uint{nbits}").dtype
        scales = ((input_high - input_low) / (levels - 1)).astype(scale_prec)
        zero_points = np.rint((levels - 1) * input_low / (input_low - input_high))
        zero_points = zero_points.astype(zero_prec)
        scale_node = make_constant(f"{ori_node.name}/QuantizeLinear_y_scale", scales)
        zp_node = make_constant(
            f"{ori_node.name}/QuantizeLinear_y_zero_point", zero_points
        )
        # add QuantizeLinear
        quantizelinear = make_node(
            "QuantizeLinear",
            inputs=[ori_node.input[0], scale_node.output[0], zp_node.output[0]],
            outputs=[f"{ori_node.name}/QuantizeLinear_output0"],
            name=f"{ori_node.name}/QuantizeLinear",
        )
        self += [scale_node, zp_node, quantizelinear]
        # replace self with DequantizeLinear
        scales = ((output_high - output_low) / (levels - 1)).astype(scale_prec)
        zero_points = np.rint((levels - 1) * output_low / (output_low - output_high))
        zero_points = zero_points.astype(zero_prec)
        scale_node = make_constant(f"{ori_node.name}/DequantizeLinear_y_scale", scales)
        zp_node = make_constant(
            f"{ori_node.name}/DequantizeLinear_y_zero_point", zero_points
        )
        self += [scale_node, zp_node]
        return make_node(
            "DequantizeLinear",
            inputs=[quantizelinear.output[0], scale_node.output[0], zp_node.output[0]],
            outputs=ori_node.output,
            name=ori_node.name,
        )



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from onnx.helper import make_graph, make_model, make_node, make_tensor_value_info

from onnxifier import ONNXIFIER_IR_VERSION, ONNXIFIER_OPSET, PassManager
from onnxifier.graph import OnnxGraph


def _build_graph():
    a = make_node("Relu", ["a"], ["a0"])
    b = make_node("Relu", ["b"], ["b0"])
    g = make_graph(
        [a, b],
        "test",
        [
            make_tensor_value_info("a", 1, [32]),
            make_tensor_value_info("b", 1, [32]),
        ],
        [
            make_tensor_value_info("a0", 1, [32]),
        ],
    )
    return make_model(
        g, ir_version=ONNXIFIER_IR_VERSION, opset_imports=[ONNXIFIER_OPSET]
    )


def test_eliminate_dead_nodes():
    model = _build_graph()
    graph = OnnxGraph(model)
    pm = PassManager(["eliminate_dead_nodes"])
    graph = pm.optimize(graph, strict=True)

    assert len(graph) == 1



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import os
from logging import Formatter, StreamHandler, addLevelName, getLevelName
from typing import Literal, Union, overload


def _default_level_from_env():
    match os.environ.get("ONNXIFIER_LOG_LEVEL"):
        case "TRACE" | "0":
            return 1
        case "DEBUG" | "1":
            return logging.DEBUG
        case "INFO" | "2":
            return logging.INFO
        case "WARNING" | "3":
            return logging.WARNING
        case "ERROR" | "4":
            return logging.ERROR
    return logging.INFO


def trace(msg: str):
    """Logging trace message"""
    _LOG.log(1, msg)


def debug(msg: str):
    """Logging debug message"""
    _LOG.debug(msg)


def info(msg: str):
    """Logging informative message"""
    _LOG.info(msg)


def warning(msg: str):
    """Logging warning message"""
    _LOG.warning(msg)


def error(msg: str):
    """Logging error message"""
    _LOG.error(msg)


def fatal(msg: str):
    """Logging critical(fatal) message"""
    _LOG.critical(msg)


def set_level(level: str):
    """Filter log messages by different level."""
    _HDL.setLevel(level.upper())


def is_enabled_for(level: Union[int, str]) -> bool:
    """Whether current log level is active."""
    if isinstance(level, str):
        from logging import _nameToLevel  # pylint:disable=import-outside-toplevel

        return _LOG.isEnabledFor(_nameToLevel[level])
    return _LOG.isEnabledFor(level)


@overload
def get_level() -> str: ...  # noqa: E704


@overload
def get_level(use_string: Literal[True]) -> str: ...  # noqa: E704


@overload
def get_level(use_string: Literal[False]) -> int: ...  # noqa: E704


def get_level(use_string: bool = True) -> Union[int, str]:
    """Get current logging level."""
    if use_string:
        return getLevelName(_HDL.level)
    return _HDL.level


def drop():
    """Drop the logger handlers and flush pending events.
    Logging after `drop` will do nothing.
    """
    for hdl in list(_LOG.handlers):
        hdl.flush()
        hdl.close()
        _LOG.removeHandler(hdl)


def nest(nested: str):
    """Create a nested logger with nested name."""
    log = logging.getLogger(f"O2O.{nested}")
    log.setLevel(_LOG.level)
    return log


addLevelName(1, "TRACE")
_LOG = logging.getLogger("O2O")
_LOG.setLevel("TRACE")
_HDL = StreamHandler()
_HDL.setLevel(_default_level_from_env())
# use colorlog to colorize different level messages
try:
    import colorlog  # pylint:disable=import-outside-toplevel

    _HDL.setFormatter(
        colorlog.ColoredFormatter(
            "[%(asctime)s]%(log_color)s[%(name)s][%(levelname)s] %(message)s"
        )
    )
except ImportError:
    _HDL.setFormatter(Formatter("[%(asctime)s][%(name)s][%(levelname)s] %(message)s"))
_LOG.addHandler(_HDL)

if os.environ.get("ONNXIFIER_LOG_FILE"):
    _FDL = logging.FileHandler(os.environ["ONNXIFIER_LOG_FILE"], encoding="utf-8")
    _FDL.setFormatter(Formatter("[%(asctime)s][%(name)s][%(levelname)s] %(message)s"))
    _FDL.setLevel(_default_level_from_env())
    _LOG.addHandler(_FDL)



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
from onnx import TensorProto
from onnx.helper import (
    make_graph,
    make_model,
    make_node,
    make_tensor,
    make_tensor_type_proto,
    make_value_info,
)

from onnxifier import ONNXIFIER_IR_VERSION, ONNXIFIER_OPSET, OnnxGraph, PassManager
from onnxifier.evaluator import Evaluator


def _build_graph():
    # Create a graph with two inputs and one output
    # The first input is quantized with scale 0.5 and zero point 0
    # The second input is not quantized
    qnode0 = make_node(
        "QuantizeLinear",
        ["input0", "scale0", "zero_point0"],
        ["quantized0"],
        "quantize0",
    )
    dqnode0 = make_node(
        "DequantizeLinear",
        ["quantized0", "scale1", "zero_point1"],
        ["output0"],
        "dequantize0",
    )
    add = make_node("Add", ["output0", "input1"], ["sum"], "add")
    graph = make_graph(
        [qnode0, dqnode0, add],
        "graph",
        [
            make_value_info("input0", make_tensor_type_proto(1, [1, 3, 224, 224])),
            make_value_info("input1", make_tensor_type_proto(1, [1, 3, 224, 224])),
        ],
        [make_value_info("sum", make_tensor_type_proto(1, [1, 3, 224, 224]))],
        [
            make_tensor("scale0", 1, [], [1.0]),
            make_tensor("zero_point0", TensorProto.UINT8, [], [0]),
            make_tensor("scale1", 1, [], [1.0]),
            make_tensor("zero_point1", TensorProto.UINT8, [], [0]),
        ],
    )
    model = make_model(
        graph, ir_version=ONNXIFIER_IR_VERSION, opset_imports=[ONNXIFIER_OPSET]
    )
    return model


def test_merge_quantize_input():
    graph = OnnxGraph(_build_graph())
    pm = PassManager(["initializer_to_constant", "infer_shape", "merge_quantize_input"])
    graph = pm.optimize(graph, strict=True)
    assert graph.tensor_type("input0") == TensorProto.UINT8
    assert graph.tensor_type("input1") == TensorProto.FLOAT

    ori_eval = Evaluator(_build_graph(), "OnnxRuntime")
    opt_eval = Evaluator(graph.model, "OnnxRuntime")

    input0 = np.random.uniform(0, 255, size=[1, 3, 224, 224]).astype(np.float32)
    input1 = np.random.uniform(-128, 127, size=[1, 3, 224, 224]).astype(np.float32)
    qinput0 = np.round(input0).clip(0, 255).astype(np.uint8)
    ori_res = ori_eval(["sum"], {"input0": input0, "input1": input1})[0]
    opt_res = opt_eval(["sum"], {"input0": qinput0, "input1": input1})[0]
    error = np.abs(ori_res - opt_res).max()
    assert np.allclose(ori_res, opt_res), f"Error: {error}"



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=missing-function-docstring

import numpy as np
import onnx
import pytest
from onnx.helper import make_graph, make_model, make_node, make_tensor_value_info
from onnx.numpy_helper import from_array

from onnxifier import ONNXIFIER_IR_VERSION, ONNXIFIER_OPSET
from onnxifier.evaluator import Evaluator
from onnxifier.graph import OnnxGraph
from onnxifier.passes.version_converter.downgrade import downgrade_op_version


def _make_resize_op19(**attrs):
    node = make_node(
        "Resize",
        ["X", "", "", "sizes"],
        ["Y"],
        "resize",
        axes=[2, 3],
        **attrs,
    )
    graph = make_graph(
        [node],
        "graph",
        [make_tensor_value_info("X", onnx.TensorProto.FLOAT, [1, 3, 24, 16])],
        [make_tensor_value_info("Y", onnx.TensorProto.FLOAT, [1, 3, "H", "W"])],
        [from_array(np.array([12, 12], np.int64), "sizes")],
    )
    model = make_model(
        graph, ir_version=ONNXIFIER_IR_VERSION, opset_imports=[ONNXIFIER_OPSET]
    )
    onnx.checker.check_model(model, True)
    return OnnxGraph(model)


def test_downgrade_op_version_api_exceptions():
    for i in range(22):
        if i in (13, 15, 17, 19):
            continue
        with pytest.raises(NotImplementedError):
            downgrade_op_version(None, i)


def test_resize_19_to_13_antialias():
    graph = _make_resize_op19(antialias=1, mode="cubic")
    with pytest.raises(ValueError):
        downgrade_op_version(graph, 13)
    graph = _make_resize_op19(antialias=1, mode="linear")
    with pytest.raises(ValueError):
        downgrade_op_version(graph, 13)


def test_resize_19_to_13():
    graph = _make_resize_op19()
    runner_ref = Evaluator(graph.model, "onnx")
    x = np.random.uniform(0, 255, size=[1, 3, 24, 16]).astype(np.float32)
    y_ref = runner_ref(["Y"], {"X": x})[0]

    graph = downgrade_op_version(graph, 13)
    onnx.checker.check_model(graph.model, True)
    runner = Evaluator(graph.model, "onnx")
    y_dut = runner(["Y"], {"X": x})[0]

    assert y_ref.shape == y_dut.shape
    assert y_ref.dtype == y_dut.dtype
    assert np.allclose(y_ref, y_dut)


def test_resize_19_to_17_keep_ratio_not_larger():
    graph = _make_resize_op19(keep_aspect_ratio_policy="not_larger")
    runner_ref = Evaluator(graph.model, "onnx")
    x = np.random.uniform(0, 255, size=[1, 3, 24, 16]).astype(np.float32)
    y_ref = runner_ref(["Y"], {"X": x})[0]

    graph = downgrade_op_version(graph, 17)
    onnx.checker.check_model(graph.model, True)
    runner = Evaluator(graph.model, "onnx")
    y_dut = runner(["Y"], {"X": x})[0]

    assert y_ref.shape == y_dut.shape
    assert y_ref.dtype == y_dut.dtype
    assert np.allclose(y_ref, y_dut)


def test_resize_19_to_17_keep_ratio_not_smaller():
    graph = _make_resize_op19(keep_aspect_ratio_policy="not_smaller")
    runner_ref = Evaluator(graph.model, "onnx")
    x = np.random.uniform(0, 255, size=[1, 3, 24, 16]).astype(np.float32)
    y_ref = runner_ref(["Y"], {"X": x})[0]

    graph = downgrade_op_version(graph, 17)
    onnx.checker.check_model(graph.model, True)
    runner = Evaluator(graph.model, "onnx")
    y_dut = runner(["Y"], {"X": x})[0]

    assert y_ref.shape == y_dut.shape
    assert y_ref.dtype == y_dut.dtype
    assert np.allclose(y_ref, y_dut)



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from onnx import TensorProto
from onnx.checker import check_model
from onnx.helper import (
    make_graph,
    make_model,
    make_node,
    make_tensor_type_proto,
    make_tensor_value_info,
    make_value_info,
)

from onnxifier import ONNXIFIER_OPSET, PassManager
from onnxifier.graph import OnnxGraph


def _build_graph_double_identities():
    identity1 = make_node("Identity", ["x"], ["y"], "identity1")
    identity2 = make_node("Identity", ["y"], ["z"], "identity2")
    relu = make_node("Relu", ["z"], ["w"], "relu")
    graph = make_graph(
        [identity1, identity2, relu],
        "graph",
        [make_value_info("x", make_tensor_type_proto(1, [1, 3, 256, 256]))],
        [make_value_info("w", make_tensor_type_proto(1, [1, 3, 256, 256]))],
        [],
    )
    return make_model(graph, opset_imports=[ONNXIFIER_OPSET])


def _build_graph_output_identity():
    relu = make_node("Relu", ["x"], ["y"], "relu")
    identity = make_node("Identity", ["y"], ["z"], "identity")
    graph = make_graph(
        [relu, identity],
        "graph",
        [make_value_info("x", make_tensor_type_proto(1, [1, 3, 256, 256]))],
        [make_value_info("z", make_tensor_type_proto(1, [1, 3, 256, 256]))],
        [],
    )
    return make_model(graph, opset_imports=[ONNXIFIER_OPSET])


def test_eliminate_double_identities():
    model = _build_graph_double_identities()
    graph = OnnxGraph(model)
    pm = PassManager(["eliminate_identity"])
    graph = pm.optimize(graph, strict=True)
    assert len(graph.nodes) == 1
    assert "identity1" not in graph
    assert "identity2" not in graph
    assert "x" in graph.inputs
    assert "w" in graph.outputs
    check_model(graph.model, True)


def test_eliminate_output_identity():
    model = _build_graph_output_identity()
    graph = OnnxGraph(model)
    pm = PassManager(["eliminate_identity"])
    graph = pm.optimize(graph, strict=True)
    assert len(graph.nodes) == 1
    assert "identity" not in graph
    assert "x" in graph.inputs
    assert "z" in graph.outputs
    check_model(graph.model, True)


def test_eliminate_identity_with_siblings():
    nodes = [
        make_node("Relu", ["x"], ["y"], "relu"),
        make_node("Identity", ["y"], ["z"], "identity"),
        make_node("Sin", ["y"], ["w"], "sin"),
    ]
    model = make_model(
        make_graph(
            nodes,
            "graph",
            [make_tensor_value_info("x", TensorProto.FLOAT, [32])],
            [
                make_tensor_value_info("z", TensorProto.FLOAT, [32]),
                make_tensor_value_info("w", TensorProto.FLOAT, [32]),
            ],
        ),
        opset_imports=[ONNXIFIER_OPSET],
    )
    graph = OnnxGraph(model)
    pm = PassManager(["eliminate_identity"])
    graph = pm.optimize(graph, strict=True)
    assert len(graph.nodes) == 2
    assert "identity" not in graph
    assert "x" in graph.inputs
    assert "z" in graph.outputs
    assert "w" in graph.outputs
    check_model(graph.model, True)



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=arguments-differ
from typing import List, Literal, Optional

from onnx.helper import make_node
from onnx.onnx_pb import NodeProto

from ... import OnnxGraph
from .. import PASSES
from ..pattern import GraphPattern, SingleNodePattern
from ..rewriter import Rewriter


@PASSES.register("split_sdpa")
class SplitSDPARewriter(Rewriter):
    """Split scaled dot-product attention (SDPA) from hidden dimension like
    FlashAttention.

    Before:

        H = softmax(Q @ K) @ V

    After:

        H = Sum(exp(Q @ K_i) @ V_i) / Sum(exp(Q @ K_i))

    Note:
        The dimensions should be static.
    """

    def __init__(self):
        pattern = GraphPattern()
        act = SingleNodePattern()
        pattern.add_edge(SingleNodePattern("MatMul"), act)
        pattern.add_edge(act, SingleNodePattern("MatMul"))
        super().__init__(pattern)

    def rewrite(
        self,
        graph: OnnxGraph,
        nodes: List[NodeProto],
        blocks: int = 64,
        mode: Optional[Literal["split_by_hidden_dims", "split_by_heads"]] = None,
    ):
        if mode == "split_by_hidden_dims":
            self.split_by_hidden_dims(graph, nodes, blocks)
        elif mode == "split_by_heads":
            self.split_by_heads(graph, nodes, 1)
        else:
            self.split_by_heads(graph, nodes, 1)

    def split_by_hidden_dims(
        self, graph: OnnxGraph, nodes: List[NodeProto], blocks: int = 64
    ):
        mm0, act, mm1 = nodes
        # query = self.get_input_node_or_die(mm0, 0)
        key = self.get_input_node_or_die(mm0, 1)
        value = self.get_input_node_or_die(mm1, 1)

        hidden_dims = graph.static_tensor_shape(key.output[0])[-1]
        if hidden_dims % blocks != 0:
            raise ValueError(
                f"Hidden dimension {hidden_dims} is not divisible by blocks {blocks}"
            )

        split0_output = [
            f"{mm0.name}/split_output{i}" for i in range(hidden_dims // blocks)
        ]
        split0 = make_node(
            "Split",
            inputs=[key.output[0]],
            outputs=split0_output,
            name=f"{mm0.name}/split",
            num_outputs=hidden_dims // blocks,
            axis=-1,
        )
        split1_output = [
            f"{mm1.name}/split_output{i}" for i in range(hidden_dims // blocks)
        ]
        split1 = make_node(
            "Split",
            inputs=[value.output[0]],
            outputs=split1_output,
            name=f"{mm1.name}/split",
            num_outputs=hidden_dims // blocks,
            axis=-2,
        )
        self += [split0, split1]
        sum_inputs = []
        for i, (key_out, value_out) in enumerate(zip(split0_output, split1_output)):
            sub_mm0 = make_node(
                "MatMul",
                inputs=[mm0.input[0], key_out],
                outputs=[f"{mm0.name}_{i}_output0"],
                name=f"{mm0.name}_{i}",
            )
            sub_act = make_node(
                act.op_type,
                inputs=[sub_mm0.output[0]],
                outputs=[f"{act.name}_{i}_output0"],
                name=f"{act.name}_{i}",
            )
            sub_act.attribute.extend(act.attribute)
            sub_mm1 = make_node(
                "MatMul",
                inputs=[sub_act.output[0], value_out],
                outputs=[f"{mm1.name}_{i}_output0"],
                name=f"{mm1.name}_{i}",
            )
            sum_inputs.append(sub_mm1.output[0])
            self += [sub_mm0, sub_act, sub_mm1]

        reduce_sum = make_node(
            "Sum", inputs=sum_inputs, outputs=[mm1.output[0]], name=f"{mm1.name}/sum"
        )

        self += reduce_sum
        self -= [mm0, mm1, act]

    def split_by_heads(
        self, graph: OnnxGraph, nodes: List[NodeProto], blocks: int = 64
    ):
        mm0, act, mm1 = nodes
        query = self.get_input_node_or_die(mm0, 0)
        key = self.get_input_node_or_die(mm0, 1)
        value = self.get_input_node_or_die(mm1, 1)

        heads = graph.static_tensor_shape(query.output[0])[1]
        if heads % blocks != 0:
            raise ValueError(f"Heads {heads} is not divisible by blocks {blocks}")

        split0a_output = [
            f"{mm0.name}/split0a_output{i}" for i in range(heads // blocks)
        ]
        split0a = make_node(
            "Split",
            inputs=[query.output[0]],
            outputs=split0a_output,
            name=f"{mm0.name}/split0a",
            num_outputs=heads // blocks,
            axis=1,
        )
        split0b_output = [
            f"{mm0.name}/split0b_output{i}" for i in range(heads // blocks)
        ]
        split0b = make_node(
            "Split",
            inputs=[key.output[0]],
            outputs=split0b_output,
            name=f"{mm0.name}/split0b",
            num_outputs=heads // blocks,
            axis=1,
        )
        split1_output = [f"{mm1.name}/split_output{i}" for i in range(heads // blocks)]
        split1 = make_node(
            "Split",
            inputs=[value.output[0]],
            outputs=split1_output,
            name=f"{mm1.name}/split",
            num_outputs=heads // blocks,
            axis=1,
        )
        self += [split0a, split0b, split1]
        sum_inputs = []
        for i, (query_out, key_out, value_out) in enumerate(
            zip(split0a_output, split0b_output, split1_output)
        ):
            sub_mm0 = make_node(
                "MatMul",
                inputs=[query_out, key_out],
                outputs=[f"{mm0.name}_{i}_output0"],
                name=f"{mm0.name}_{i}",
            )
            sub_act = make_node(
                act.op_type,
                inputs=[sub_mm0.output[0]],
                outputs=[f"{act.name}_{i}_output0"],
                name=f"{act.name}_{i}",
            )
            sub_act.attribute.extend(act.attribute)
            sub_mm1 = make_node(
                "MatMul",
                inputs=[sub_act.output[0], value_out],
                outputs=[f"{mm1.name}_{i}_output0"],
                name=f"{mm1.name}_{i}",
            )
            sum_inputs.append(sub_mm1.output[0])
            self += [sub_mm0, sub_act, sub_mm1]

        concat = make_node(
            "Concat",
            inputs=sum_inputs,
            outputs=[mm1.output[0]],
            axis=1,
            name=f"{mm1.name}/concat",
        )

        self += concat
        self -= [mm0, mm1, act]



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List

import numpy as np
from onnx import NodeProto
from onnx.helper import make_node

from ... import OnnxGraph
from .. import PASSES
from ..pattern import SingleNodePattern
from ..rewriter import Rewriter
from ..utils import make_constant


@PASSES.register("col2im_to_depthtospace")
class Col2ImToDepthToSpaceRewriter(Rewriter):
    """Convert simple Col2Im to DepthToSpace.

    Col2Im must satisfy strides == block_size

    Before:

        Col2Im

    After:

        Reshape -> DepthToSpace
    """

    def __init__(self):
        super().__init__(SingleNodePattern("Col2Im"))

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto]):
        node = nodes[0]

        dilations = self.get_attribute(node, "dilations", [1])
        if isinstance(dilations, list) and any(i != 1 for i in dilations):
            return

        strides = self.get_attribute(node, "strides")
        if not isinstance(strides, list):
            return

        image_shape = self.get_value_or_die(node.input[1])
        block_size = self.get_value_or_die(node.input[2])
        input_shape = graph.tensor_shape(node.input[0])

        if (block_size != strides).any() or any(i != block_size[0] for i in block_size):
            return

        b, c, _ = input_shape
        output_shape = [b, c, *[i // s for i, s in zip(image_shape, strides)]]
        shape_cst = make_constant(
            f"{node.name}/shape", np.array(output_shape, dtype=np.int64)
        )
        reshape = make_node(
            "Reshape",
            [node.input[0], shape_cst.output[0]],
            [f"{node.name}/Reshape_output0"],
            name=f"{node.name}/Reshape",
        )
        d2s = make_node(
            "DepthToSpace",
            [reshape.output[0]],
            [node.output[0]],
            name=node.name,
            blocksize=int(block_size[0]),
            mode="CRD",
        )

        self += [shape_cst, reshape, d2s]
        self -= node



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import onnx
from onnx.helper import (
    make_attribute,
    make_graph,
    make_model,
    make_node,
    make_tensor,
    make_tensor_type_proto,
    make_tensor_value_info,
    make_value_info,
)

from onnxifier import OnnxGraph
from onnxifier.passes.pattern import (
    GraphPattern,
    InputNodePattern,
    OutputNodePattern,
    SingleNodePattern,
    StartEndPointPattern,
)


def _build_test_graph1():
    conv0 = make_node(
        "Conv",
        inputs=["a", "w0"],
        outputs=["c"],
        group=2,
        name="conv0",
        domain="test.domain",
    )
    conv1 = make_node("Conv", inputs=["c", "w1"], outputs=["d"], group=1, name="conv1")
    graph = make_graph(
        [conv0, conv1],
        name="graph",
        inputs=[make_value_info("a", make_tensor_type_proto(1, [1, 3, 128, 127]))],
        outputs=[make_value_info("d", make_tensor_type_proto(1, None))],
        initializer=[
            make_tensor(
                "w0",
                1,
                [8, 3, 3, 3],
                np.ones([8, 3, 3, 3], "float32").tobytes(),
                raw=True,
            ),
            make_tensor(
                "w1",
                1,
                [8, 8, 3, 3],
                np.ones([8, 8, 3, 3], "float32").tobytes(),
                raw=True,
            ),
        ],
    )
    model = make_model(graph)
    return model


def _build_test_graph2():
    conv0 = make_node("Conv", inputs=["a", "w0"], outputs=["c"], group=2, name="conv0")
    conv1 = make_node("Conv", inputs=["c", "w1"], outputs=["d"], group=1, name="conv1")
    add = make_node("Add", inputs=["c", "d"], outputs=["y"], name="add")
    graph = make_graph(
        [conv0, conv1, add],
        name="graph",
        inputs=[make_value_info("a", make_tensor_type_proto(1, [1, 3, 128, 128]))],
        outputs=[make_value_info("y", make_tensor_type_proto(1, None))],
        initializer=[
            make_tensor(
                "w0",
                1,
                [8, 3, 3, 3],
                np.ones([8, 3, 3, 3], "float32").tobytes(),
                raw=True,
            ),
            make_tensor(
                "w1",
                1,
                [8, 8, 3, 3],
                np.ones([8, 8, 3, 3], "float32").tobytes(),
                raw=True,
            ),
        ],
    )
    model = make_model(graph)
    return model


def test_single_node_match():
    graph = OnnxGraph(_build_test_graph1())
    pattern = SingleNodePattern("Conv")
    nodes = list(pattern.match(graph))
    assert len(nodes) == 2
    for i in nodes:
        assert isinstance(i, onnx.NodeProto)
        assert i.op_type == "Conv"

    pattern = SingleNodePattern("Conv").with_attr("group", 1)
    nodes = list(pattern.match(graph))
    assert len(nodes) == 1
    for i in nodes:
        assert i.attribute[0].i == 1

    pattern1 = SingleNodePattern("Conv").with_attr(make_attribute("group", 1))
    pattern2 = SingleNodePattern("Conv").with_attr("group", 2)
    nodes = list((pattern1 | pattern2).match(graph))
    assert len(nodes) == 2
    for i in nodes:
        assert isinstance(i, onnx.NodeProto)
        assert i.op_type == "Conv"
        assert i.attribute[0].i in (1, 2)

    pattern = SingleNodePattern("Conv").with_attr("dilations")
    nodes = list(pattern.match(graph))
    assert len(nodes) == 0

    pattern = SingleNodePattern().with_name("conv1")
    nodes = list(pattern.match(graph))
    assert len(nodes) == 1
    assert nodes[0].op_type == "Conv"

    pattern = SingleNodePattern("Conv").with_domain("test.domain")
    nodes = list(pattern.match(graph))
    assert len(nodes) == 1

    pattern = SingleNodePattern("Conv").with_domain("*")
    nodes = list(pattern.match(graph))
    assert len(nodes) == 1

    pattern = SingleNodePattern("Conv")
    nodes = list(pattern.match(graph))
    assert len(nodes) == 2


def test_single_node_match_inputs_outputs():
    graph = OnnxGraph(_build_test_graph2())
    pattern = SingleNodePattern("Conv").with_inputs(2)
    nodes = list(pattern.match(graph))
    assert len(nodes) == 1
    assert nodes[0].op_type == "Conv"
    assert nodes[0].name == "conv1"

    pattern = SingleNodePattern("Conv").with_inputs(2, "conv0", "w1")
    nodes = list(pattern.match(graph))
    assert len(nodes) == 1
    assert nodes[0].op_type == "Conv"
    assert nodes[0].name == "conv1"

    pattern = SingleNodePattern("Conv").with_outputs(2, "add", None)
    nodes = list(pattern.match(graph))
    assert len(nodes) == 1
    assert nodes[0].op_type == "Conv"
    assert nodes[0].name == "conv0"

    pattern = SingleNodePattern().with_outputs(1)
    nodes = list(pattern.match(graph))
    assert len(nodes) == 1
    assert nodes[0].op_type == "Conv"
    assert nodes[0].name == "conv1"


def test_single_node_match_order():
    g = make_graph(
        [
            make_node("Relu", ["1"], ["2"], name="relu1"),
            make_node("Relu", ["2"], ["3"], name="relu2"),
            make_node("Relu", ["3"], ["4"], name="relu3"),
        ],
        name="graph",
        inputs=[make_tensor_value_info("1", onnx.TensorProto.FLOAT, [10])],
        outputs=[make_tensor_value_info("4", onnx.TensorProto.FLOAT, [10])],
    )
    model = make_model(g)
    graph = OnnxGraph(model)

    pattern = SingleNodePattern("Relu").with_order("post")
    nodes = list(pattern.match(graph))
    assert nodes[0].name == g.node[2].name
    assert nodes[1].name == g.node[1].name
    assert nodes[2].name == g.node[0].name

    pattern = SingleNodePattern("Relu").with_order("pre")
    nodes = list(pattern.match(graph))
    assert nodes[0].name == g.node[0].name
    assert nodes[1].name == g.node[1].name
    assert nodes[2].name == g.node[2].name


def test_or_pattern():
    graph = OnnxGraph(_build_test_graph1())
    pattern1 = SingleNodePattern("Conv").with_attr("group", 1)
    pattern2 = SingleNodePattern("Conv").with_attr("group", 2)
    nodes = list((pattern1 | pattern2).match(graph))
    assert len(nodes) == 2
    nodes = list((pattern1 + pattern2).match(graph))
    assert len(nodes) == 2
    for i in nodes:
        assert isinstance(i, onnx.NodeProto)
        assert i.op_type == "Conv"
        assert i.attribute[0].i in (1, 2)


def test_subgraph_match():
    graph = OnnxGraph(_build_test_graph1())
    pattern = GraphPattern().add_edge(
        SingleNodePattern("Conv"), SingleNodePattern("Conv")
    )
    nodes = list(pattern.match(graph))
    assert len(nodes) == 1
    assert len(nodes[0]) == 2

    pattern = GraphPattern(pattern)
    nodes = list(pattern.match(graph))
    assert len(nodes) == 1
    assert len(nodes[0]) == 2


def test_input_node_match():
    graph = OnnxGraph(_build_test_graph1())
    pattern = InputNodePattern()
    nodes = list(pattern.match(graph))
    assert len(nodes) == 1
    assert nodes[0].name == "conv0"


def test_output_node_match():
    graph = OnnxGraph(_build_test_graph1())
    pattern = OutputNodePattern()
    nodes = list(pattern.match(graph))
    assert len(nodes) == 1
    assert nodes[0].name == "conv1"


def test_start_end_point_match():
    graph = OnnxGraph(_build_test_graph2())
    pattern = StartEndPointPattern(
        SingleNodePattern(op_name="conv0"), SingleNodePattern(op_name="add")
    )
    nodes_list = list(pattern.match(graph))
    assert len(nodes_list) == 1
    for nodes in nodes_list:
        assert any(i.name == "conv1" for i in nodes)

    pattern_a = GraphPattern()
    pattern_a.add_edge(SingleNodePattern("Conv"), SingleNodePattern("Conv"))
    pattern = StartEndPointPattern(pattern_a, SingleNodePattern("Add"))
    nodes_list = list(pattern.match(graph))
    assert len(nodes_list) == 1
    assert len(nodes_list[0]) == 3

    pattern_b = GraphPattern()
    pattern_b.add_edge(
        SingleNodePattern().with_name("conv1"), SingleNodePattern().with_name("add")
    )
    pattern = StartEndPointPattern(SingleNodePattern("Conv"), pattern_b)
    nodes_list = list(pattern.match(graph))
    assert len(nodes_list) == 1
    assert len(nodes_list[0]) == 3


def test_single_match_pattern_with_specify_node_names():
    graph = OnnxGraph(_build_test_graph2())
    pattern = SingleNodePattern()
    nodes = list(pattern.match(graph, specify_node_names={"add", "conv1"}))
    assert len(nodes) == 2
    assert {n.name for n in nodes} == {"add", "conv1"}


def test_graph_pattern_with_specify_node_names():
    graph = OnnxGraph(_build_test_graph2())
    pattern = GraphPattern().add_edge(
        SingleNodePattern("Conv"), SingleNodePattern("Conv")
    )
    nodes = list(pattern.match(graph, specify_node_names={"conv0"}))
    assert len(nodes) == 1
    assert {n.name for n in nodes[0]} == {"conv0", "conv1"}

    nodes = list(pattern.match(graph, specify_node_names={"conv1"}))
    assert len(nodes) == 1
    assert {n.name for n in nodes[0]} == {"conv0", "conv1"}

    nodes = list(pattern.match(graph, specify_node_names={"conv2"}))
    assert len(nodes) == 0


def test_input_node_pattern_with_specify_node_names():
    graph = OnnxGraph(_build_test_graph2())
    pattern = InputNodePattern()
    nodes = list(pattern.match(graph, specify_node_names={"conv0"}))
    assert len(nodes) == 1
    assert nodes[0].name == "conv0"

    nodes = list(pattern.match(graph, specify_node_names={"conv1"}))
    assert len(nodes) == 0


def test_output_node_pattern_with_specify_node_names():
    graph = OnnxGraph(_build_test_graph2())
    pattern = OutputNodePattern()
    nodes = list(pattern.match(graph, specify_node_names={"add"}))
    assert len(nodes) == 1
    assert nodes[0].name == "add"

    nodes = list(pattern.match(graph, specify_node_names={"conv1"}))
    assert len(nodes) == 0



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List

import numpy as np
import onnx

from ... import OnnxGraph, logger
from .. import L1
from ..pattern import SingleNodePattern
from ..rewriter import Rewriter
from ..utils import make_constant


@L1.register(name="scatter_nd_indices_to_positive", deps=["infer_shape"])
class ScatterNDIndicesToPositiveRewriter(Rewriter):
    r"""This pass converts potential negative indices in ScatterND op to
    positive indices.

    Before:

        ScatterND(data, indices, updates)

    After:

        ScatterND(data, indices_pos, updates)
    """

    def __init__(self):
        super().__init__(SingleNodePattern("ScatterND"))

    def rewrite(self, graph: OnnxGraph, nodes: List[onnx.NodeProto]):
        node = nodes[0]
        indices = self.get_value_or_die(node.input[1]).copy()
        data_shape = graph.static_tensor_shape(node.input[0])
        if np.all(indices >= 0):  # no negative indices
            return

        logger.debug(f"Found negative indices in ScatterND {node.name}")
        logger.debug(f"Data shape: {data_shape}")

        for i, d in enumerate(data_shape):
            indices[..., i] = np.where(
                indices[..., i] < 0, indices[..., i] + d, indices[..., i]
            )
        pos_indices = make_constant(f"{node.name}/indices_pos", indices)
        node.input[1] = pos_indices.output[0]
        self += pos_indices



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import onnx
from onnx.helper import make_graph, make_model, make_node, make_tensor_value_info

from onnxifier import ONNXIFIER_IR_VERSION, ONNXIFIER_OPSET, OnnxGraph
from onnxifier.algo.subgraph import find_sibling_nodes


def _make_graph():
    sin1 = make_node("Sin", ["X1"], ["S1"], name="sin1")
    sin2 = make_node("Sin", ["S1"], ["S2"], name="sin2")
    act2 = make_node("Relu", ["S2"], ["A2"], name="act2")
    sin3 = make_node("Sin", ["S1"], ["S3"], name="sin3")
    act3 = make_node("Relu", ["S3"], ["A3"], name="act3")

    graph = make_graph(
        [sin1, sin2, act2, sin3, act3],
        "test_graph",
        [
            make_tensor_value_info("X1", onnx.TensorProto.FLOAT, [1, 2, 3, 4]),
        ],
        [
            make_tensor_value_info("A2", onnx.TensorProto.FLOAT, [1, 2, 3, 4]),
            make_tensor_value_info("A3", onnx.TensorProto.FLOAT, [1, 2, 3, 4]),
        ],
    )
    return make_model(
        graph, ir_version=ONNXIFIER_IR_VERSION, opset_imports=[ONNXIFIER_OPSET]
    )


def test_find_sibling_nodes():
    graph = OnnxGraph(_make_graph())

    siblings = find_sibling_nodes(graph, graph.nodes["sin2"]["pb"], parents=["Sin"])
    assert len(siblings) == 1
    assert siblings[0].name == "sin3"



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List

import numpy as np
from onnx.helper import make_node
from onnx.onnx_pb import NodeProto

from ... import OnnxGraph
from .. import PASSES
from ..pattern import GraphPattern, SingleNodePattern
from ..rewriter import Rewriter
from ..utils import make_constant

ACT_SHAVE = (
    "Sigmoid",
    "Swish",
    "Mish",
    "Gelu",
    "Tanh",
)


@PASSES.register(
    "insert_conv_before_act_shave",
    deps=["fuse_mish", "fuse_swish"],
)
class InsertConvBeforeActShaveRewriter(Rewriter):
    def __init__(self):
        pattern = GraphPattern()
        patterns = [SingleNodePattern(act) for act in ACT_SHAVE]
        # TODO: quantized?
        pattern.add_edge(sum(patterns), SingleNodePattern("Conv"))
        super().__init__(pattern=pattern)

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto], *args, **kwargs):
        act_node, conv_node = nodes
        if len(graph.onnx_successors(act_node)) > 1:
            return  # make sure act has only one conv user

        _, channels, _, _ = graph.tensor_shape(act_node.output[0])
        assert isinstance(channels, int) and channels > 1
        weight1 = make_constant(
            f"{act_node.name}/conv1/weight",
            np.ones((channels, 1, 1, 1), np.float32),
        )
        conv1 = make_node(
            "Conv",
            [act_node.output[0], weight1.output[0]],
            [f"{act_node.name}/conv1_output0"],
            name=f"{act_node.name}/conv1",
            group=channels,
        )
        conv_node.input[0] = conv1.output[0]
        self += [weight1, conv1]



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=arguments-differ
from typing import List

import numpy as np
from onnx.helper import make_node, tensor_dtype_to_np_dtype
from onnx.onnx_pb import NodeProto

from ... import OnnxGraph
from .. import PASSES
from ..pattern import GraphPattern, SingleNodePattern
from ..rewriter import Rewriter
from ..utils import make_constant


@PASSES.register(name="space2depth_to_qconv", deps=["initializer_to_constant"])
class SpaceToDepthToQConvRewriter(Rewriter):
    """Convert SpaceToDepth - QDQ to quantized Conv."""

    def __init__(self):
        qdq_pattern = GraphPattern()
        s2d = SingleNodePattern("SpaceToDepth")
        qpattern = SingleNodePattern("QuantizeLinear")
        dqpattern = SingleNodePattern("DequantizeLinear")
        qdq_pattern.add_edge(s2d, qpattern)
        qdq_pattern.add_edge(qpattern, dqpattern)
        super().__init__(pattern=qdq_pattern)

    def _copy_qdq(self, node: NodeProto) -> NodeProto:
        axis = self.get_attribute(node, "axis") or 0
        scale_node = self.get_input_node_or_die(node, 1)
        scale_value = self.get_value_or_die(scale_node)
        new_scale = make_constant(node.name + "/scale/copy", scale_value)
        input_names = [node.input[0] + "/copy", new_scale.output[0]]
        if len(node.input) == 3:
            zero_node = self.get_input_node_or_die(node, 2)
            zero_value = self.get_value_or_die(zero_node)
            new_zero = make_constant(node.name + "/zero/copy", zero_value)
            self += new_zero
            input_names += new_zero.output[:]
        new_node = make_node(
            node.op_type,
            input_names,
            [o + "/copy" for o in node.output],
            node.name + "/s2d_to_qconv/copy",
            axis=axis,
        )
        self += [new_scale]
        return new_node

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto]):
        s2d, qnode, dqnode = nodes
        blocksize = self.get_attribute(s2d, "blocksize")
        assert isinstance(blocksize, int)
        _, ic, _, _ = graph.tensor_shape(s2d.input[0])
        _, oc, _, _ = graph.tensor_shape(s2d.output[0])
        assert isinstance(ic, int) and isinstance(oc, int)
        etype = graph.tensor_type(dqnode.input[0])
        dtype = tensor_dtype_to_np_dtype(etype)
        assert ic * blocksize**2 == oc, "invalid space2depth parameters"

        qnode_ahead = self._copy_qdq(qnode)
        dqnode_ahead = self._copy_qdq(dqnode)
        qnode_ahead.input[0] = s2d.input[0]
        dqnode_ahead.input[0] = qnode_ahead.output[0]

        kernel = (
            # fmt: off
            np.tile(
                np.eye(blocksize**2)
                .reshape([-1, blocksize, blocksize])
                .astype(dtype),
                [ic, 1, 1, 1],
            )
            .transpose([1, 0, 2, 3])
            .reshape([oc, 1, blocksize, blocksize])
            # fmt: on
        )
        # expand to [oc, ic, blocksize, blocksize]
        kernel = np.concatenate(
            [kernel, np.tile(np.zeros_like(kernel), [1, ic - 1, 1, 1])], axis=1
        )
        for i in range(1, ic):
            kernel[i::ic][:, i] = kernel[i::ic][:, 0]
            kernel[i::ic][:, 0] = 0  # swap
        kernel_max = np.iinfo(dtype).max
        conv_weight = make_constant(
            name=f"{s2d.name}/weight", value=kernel * kernel_max
        )

        # kernel dequantize
        kernel_scale = make_constant(
            conv_weight.name + "/scale", np.ones([oc], np.float32) / kernel_max
        )
        kernel_zp = make_constant(conv_weight.name + "/zero", np.zeros([oc], dtype))
        kernel_dq = make_node(
            "DequantizeLinear",
            [conv_weight.output[0], kernel_scale.output[0], kernel_zp.output[0]],
            [conv_weight.output[0] + "/dq"],
            conv_weight.name + "/dequantize",
            axis=0,
        )

        conv = make_node(
            op_type="Conv",
            inputs=dqnode_ahead.output[:] + [kernel_dq.output[0]],
            outputs=s2d.output,
            name=f"{s2d.name}/conv",
            kernel_shape=[blocksize, blocksize],
            strides=[blocksize, blocksize],
        )
        self -= s2d
        self += [
            qnode_ahead,
            dqnode_ahead,
            kernel_dq,
            kernel_scale,
            kernel_zp,
            conv_weight,
            conv,
        ]



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List

import networkx as nx
from onnx import NodeProto

from ... import OnnxGraph
from ...evaluator import Evaluator
from .. import L1
from ..pattern import ConstantGraphPattern
from ..rewriter import Rewriter
from ..utils import make_constant


@L1.register(
    name="fold_constant",
    deps=["initializer_to_constant", "infer_shape", "shape_to_constant"],
)
class FoldConstantPass(Rewriter):
    """Fold constants to a single node."""

    def __init__(self):
        super().__init__(pattern=ConstantGraphPattern())

    def _is_constant_or_qdq(self, node):
        return node.op_type in {"Constant", "DequantizeLinear", "QuantizeLinear"}

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto]):
        # skip if all nodes are Constant
        if all(self._is_constant_or_qdq(node) for node in nodes):
            return
        subonnx = graph.onnx_subgraph(nodes)
        out_nodes = {}
        constants = []
        for node_name in nx.topological_sort(subonnx):
            node_pb = subonnx.nodes[node_name]["pb"]
            if subonnx.nodes[node_name]["has_output"]:
                for out_name in node_pb.output:
                    if out_name in subonnx.outputs:
                        out_nodes[out_name] = node_pb
        # filter out the outputs that are exactly constant's output
        outputs_eval = set(subonnx.outputs)
        for node in list(filter(lambda n: n.op_type == "Constant", nodes)):
            if set(node.output).issubset(outputs_eval):
                nodes.remove(node)
                outputs_eval.difference_update(node.output)
        evaluator = Evaluator(subonnx.model)
        outputs = evaluator(list(outputs_eval), {})
        for output_name, out_value in zip(outputs_eval, outputs):
            constants.append(make_constant(output_name, out_value))
            constants[-1].output[0] = output_name  # keep the edge name
        self -= nodes
        self += constants
